from ingenialink.canopen.register import CanopenRegister
from ingenialink.enums.register import (
    REG_ACCESS,
    REG_ADDRESS_TYPE,
    REG_DTYPE,
    REG_PHY,
    RegCyclicType,
)


class EthercatRegister(CanopenRegister):
    pass



import os
import pathlib
import platform
import socket
import time
from enum import Enum, IntEnum
from threading import Thread, Timer
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from scipy import signal

from ingenialink.constants import ETH_BUF_SIZE, MONITORING_BUFFER_SIZE
from ingenialink.enums.register import REG_ACCESS, REG_DTYPE
from ingenialink.ethernet.register import EthernetRegister
from ingenialink.ethernet.servo import EthernetServo
from ingenialink.utils import constants
from ingenialink.utils._utils import convert_bytes_to_dtype, convert_dtype_to_bytes
from ingenialink.utils.constants import IL_MC_CW_EO
from ingenialink.utils.mcb import MCB
from ingenialink.virtual.dictionary import VirtualDictionary

R_VALUE = 1.1
L_VALUE = 3.9e-4
J_VALUE = 0.012
B_VALUE = 0.027
TORQUE_CONSTANT = 0.05

STATUS_WORD_COMMUTATION_FEEDBACK_ALIGNED_BIT = 0x4000

INC_ENC1_RESOLUTION = 4096
INC_ENC2_RESOLUTION = 2000


class MSG_TYPE(Enum):
    RECEIVED = "RECEIVED"
    SENT = "SENT"


class OperationMode(IntEnum):
    """Operation Mode Enum"""

    VOLTAGE = 0x00
    CURRENT_AMPLIFIER = 0x01
    CURRENT = 0x02
    CYCLIC_CURRENT = 0x22
    VELOCITY = 0x03
    PROFILE_VELOCITY = 0x13
    CYCLIC_VELOCITY = 0x23
    POSITION = 0x04
    PROFILE_POSITION = 0x14
    CYCLIC_POSITION = 0x24
    PROFILE_POSITION_S_CURVE = 0x44
    INTERPOLATED_POSITION = 0xA4
    PVT = 0xB4
    HOMING = 0x113
    TORQUE = 0x05
    CYCLIC_TORQUE = 0x25


class PhasingMode(IntEnum):
    """Phasing modes"""

    NON_FORCED = 0
    """Non forced"""
    FORCED = 1
    """Forced"""
    NO_PHASING = 2
    """No phasing"""


class SensorType(IntEnum):
    """Summit series feedback type enum"""

    ABS1 = 1
    """Absolute encoder 1"""
    INTGEN = 3
    """Internal generator"""
    QEI = 4
    """Digital/Incremental encoder 1"""
    HALLS = 5
    """Digital halls"""
    SSI2 = 6
    """Secondary SSI"""
    BISSC2 = 7
    """Absolute encoder 2"""
    QEI2 = 8
    """Digital/Incremental encoder 2"""


class GeneratorMode(IntEnum):
    """Generator modes"""

    CONSTANT = 0
    """Constant"""
    SAW_TOOTH = 1
    """Saw tooth"""
    SQUARE = 2
    """Square"""


class BasePlant:
    """Base class for open-loop and closed-loop plants.

    Args:
        drive: Instance of VirtualDrive.

    Attributes:
        drive: Instance of VirtualDrive.
        plant: Plant's discrete transfer function.
        monitoring_frequency: Monitoring frequency
        set_point_register: Set-point register.
        command_register: Command register.
        value_register: Value register.
    """

    REGISTER_SET_POINT: str
    """Register of the set-point signal."""
    REGISTER_COMMAND: str
    """Register of the command signal."""
    REGISTER_VALUE: str
    """Register of the value (output) signal."""

    def __init__(self, drive: "VirtualDrive") -> None:
        self.drive = drive
        self.monitoring_frequency = VirtualMonitoring.FREQUENCY
        self.plant: signal.TransferFunction
        self.set_point_register = self.drive.get_register(1, id=self.REGISTER_SET_POINT)
        self.command_register = self.drive.get_register(1, id=self.REGISTER_COMMAND)
        self.value_register = self.drive.get_register(1, id=self.REGISTER_VALUE)

    def emulate_plant(self, from_disturbance: bool = True) -> None:
        """Emulate the plant by filtering the excitation signal with the plant's frequency response.

        Args:
            from_disturbance: If True the input signal is repeated to fit the monitoring window.
        """
        dist_signal = self.drive.reg_signals[self.REGISTER_SET_POINT]
        if len(dist_signal) == 0:
            return

        if from_disturbance and self.drive._monitoring:
            monitoring_size = self.drive._monitoring.buffer_size * self.drive._monitoring.divider
            if monitoring_size > len(dist_signal):
                repetitions = monitoring_size // len(dist_signal) + 1
                dist_signal = np.tile(dist_signal, repetitions)

        initial_value = (
            self.set_point_register.storage if self.set_point_register.storage > 0 else 0
        )
        use_fft_method = self.value_register.dtype == REG_DTYPE.FLOAT
        mon_signal = self._filter_signal(
            dist_signal, use_fft_method=use_fft_method, initial_value=initial_value
        )

        self.drive.reg_signals[self.REGISTER_COMMAND] = dist_signal
        self.drive.reg_time[self.REGISTER_COMMAND] = self.drive.reg_time[self.REGISTER_SET_POINT]
        self.drive.reg_signals[self.REGISTER_VALUE] = mon_signal
        self.drive.reg_time[self.REGISTER_VALUE] = self.drive.reg_time[self.REGISTER_SET_POINT]

    def _filter_signal(
        self,
        input_signal: NDArray[np.float_],
        use_fft_method: bool = True,
        initial_value: float = 0.0,
        plant: Optional[signal.TransferFunction] = None,
    ) -> NDArray[np.float_]:
        """Filter signal with the plant's frequency response.

        Args:
            input_signal: Signal to be filtered.
            use_fft_method: If True the FFT method is used.
            initial_value: Initial value.

        Returns:
            Filtered signal.
        """
        plant = plant or self.plant
        if len(plant.num) == 1 and len(plant.den) == 1:
            gain = plant.num[0] / plant.den[0]
            output_signal = input_signal * gain
        if use_fft_method:
            input_signal_fft = np.fft.fft(input_signal)
            _, freq_response = signal.freqz(
                plant.num, plant.den, worN=len(input_signal), whole=True
            )
            output_signal_fft = input_signal_fft * freq_response
            output_signal = np.real(np.fft.ifft(output_signal_fft))
        else:
            initial_x = [initial_value] * (len(plant.num) - 1)
            initial_y = [initial_value] * (len(plant.den) - 1)
            zi = signal.lfiltic(plant.num, plant.den, initial_y, x=initial_x)
            output_signal, _ = signal.lfilter(plant.num, plant.den, input_signal, zi=zi)

        return output_signal

    def jog(self, new_value: Union[int, float]) -> None:
        """Emulate jogs by disturbing the input register using a step signal.

        Args:
            new_value: Target value.
        """
        if not self.drive.enabled:
            return
        time_of_change = time.time()
        start_time = time_of_change - 1
        end_time = start_time + 5
        time_vector = np.arange(start_time, end_time, 1 / self.drive.current_loop_rate)
        index_change = int(np.argmin(np.abs(time_vector - time_of_change)))
        jog_signal = np.zeros(len(time_vector))
        current_value = self.set_point_register.storage
        jog_signal[:index_change] = current_value
        jog_signal[index_change:] = new_value
        self.drive.reg_signals[self.REGISTER_SET_POINT] = jog_signal
        self.drive.reg_time[self.REGISTER_SET_POINT] = time_vector
        self.emulate_plant(from_disturbance=False)

    def clean_signals(self) -> None:
        """Clean all signals."""
        self.drive.reg_time[self.REGISTER_COMMAND] = np.array([])
        self.drive.reg_signals[self.REGISTER_COMMAND] = np.array([])
        self.drive.reg_time[self.REGISTER_VALUE] = np.array([])
        self.drive.reg_signals[self.REGISTER_VALUE] = np.array([])
        self.drive.reg_time[self.REGISTER_SET_POINT] = np.array([])
        self.drive.reg_signals[self.REGISTER_SET_POINT] = np.array([])


class BaseOpenLoopPlant(BasePlant):
    """Base class for open-loop plants.

    Args:
        drive: Instance of VirtualDrive.
        num: Numerator of the continuos transfer function.
        den: Denominator of the continuos transfer function.

    Attributes:
        drive: Instance of VirtualDrive.
        plant: Plant's discrete transfer function.
    """

    def __init__(self, drive: "VirtualDrive", num: List[float], den: List[float]) -> None:
        super().__init__(drive)
        self.plant = signal.TransferFunction(num, den).to_discrete(dt=1 / self.monitoring_frequency)
        self.drive.reg_noise_amplitude[self.REGISTER_VALUE] = 0.01


class BaseClosedLoopPlant(BasePlant):
    """Base class for closed-loop plants.

    Args:
        drive: Instance of VirtualDrive.
        open_loop_plant: Open-loop plant's discrete transfer function.

    Attributes:
        drive: Instance of VirtualDrive.
        open_loop_plant: Open-loop plant's discrete transfer function.
        plant: Closed-loop plant's discrete transfer function.
    """

    KI_REG: str
    """Register containing the Ki value."""
    KP_REG: str
    """Register containing the Kp value."""

    TUNING_OPERATION_MODE: Optional[OperationMode] = None
    """Operation mode used for tuning."""
    OL_PLANT_INPUT: str
    """Register of the open-loop plant's input."""

    def __init__(self, drive: "VirtualDrive", open_loop_plant: signal.TransferFunction) -> None:
        super().__init__(drive)
        self.open_loop_plant = open_loop_plant
        self.create_closed_loop_plant()

    def create_closed_loop_plant(self) -> None:
        """Create the closed-loop plant."""
        discrete_controller = signal.TransferFunction(
            [self.kp + self.kp * self.ki * 1 / self.monitoring_frequency, -self.kp],
            [1, -1],
            dt=1 / self.monitoring_frequency,
        )

        num = np.polymul(self.open_loop_plant.num, discrete_controller.num)
        den = np.polymul(self.open_loop_plant.den, discrete_controller.den)
        open_loop_control = signal.TransferFunction(num, den, dt=1 / self.monitoring_frequency)

        feedback = signal.TransferFunction([1], [1], dt=1 / self.monitoring_frequency)

        num = np.polymul(open_loop_control.num, feedback.den)
        den = np.polyadd(
            np.polymul(open_loop_control.num, feedback.num),
            np.polymul(open_loop_control.den, feedback.den),
        )

        self.plant = signal.TransferFunction(num, den, dt=1 / self.monitoring_frequency)

    def emulate_plant(self, from_disturbance: bool = True) -> None:
        self.create_closed_loop_plant()
        super().emulate_plant(from_disturbance=from_disturbance)
        self.__obtain_command_signal()

    def __obtain_command_signal(self) -> None:
        """Obtain command signal from the output by applying the inverse of the open-loop plant."""
        if not (
            self.TUNING_OPERATION_MODE is not None
            and self.drive.operation_mode == self.TUNING_OPERATION_MODE
            and len(self.drive.reg_signals[self.REGISTER_VALUE]) > 0
        ):
            return
        output_signal = self.drive.reg_signals[self.REGISTER_VALUE]
        inverse_open_loop = signal.TransferFunction(
            self.open_loop_plant.den, self.open_loop_plant.num, dt=self.open_loop_plant.dt
        )
        command_value = self._filter_signal(
            output_signal,
            use_fft_method=False,
            initial_value=output_signal[0],
            plant=inverse_open_loop,
        )
        self.drive.reg_signals[self.OL_PLANT_INPUT] = command_value
        self.drive.reg_time[self.OL_PLANT_INPUT] = self.drive.reg_time[self.REGISTER_VALUE]
        self.drive.reg_noise_amplitude[self.OL_PLANT_INPUT] = self.drive.reg_noise_amplitude[
            self.REGISTER_VALUE
        ]

    @property
    def kp(self) -> float:
        """Return Kp value.

        Returns:
            Kp value.
        """
        return float(self.drive.get_value_by_id(1, self.KP_REG))

    @property
    def ki(self) -> float:
        """Return Ki value.

        Returns:
            Ki value.
        """
        return float(self.drive.get_value_by_id(1, self.KI_REG))


class PlantOpenLoopRL(BaseOpenLoopPlant):
    """Emulator of a open-loop RL plant."""

    REGISTER_SET_POINT = "CL_VOL_D_SET_POINT"
    REGISTER_COMMAND = "CL_VOL_D_CMD"
    REGISTER_VALUE = "CL_CUR_D_VALUE"

    def __init__(self, drive: "VirtualDrive") -> None:
        self.l_henry = L_VALUE
        self.r_ohm = R_VALUE
        super().__init__(drive, [1], [self.l_henry, self.r_ohm])


class PlantClosedLoopRL(BaseClosedLoopPlant):
    """Emulator of a closed-loop RL plant."""

    REGISTER_SET_POINT = "CL_CUR_D_SET_POINT"
    REGISTER_COMMAND = "CL_CUR_D_REF_VALUE"
    REGISTER_VALUE = "CL_CUR_D_VALUE"

    KI_REG = "CL_CUR_D_KI"
    KP_REG = "CL_CUR_D_KP"

    TUNING_OPERATION_MODE = OperationMode.CURRENT
    OL_PLANT_INPUT = PlantOpenLoopRL.REGISTER_COMMAND

    def emulate_plant(self, from_disturbance: bool = True) -> None:
        super().emulate_plant(from_disturbance)


class PlantOpenLoopRLQuadrature(BaseOpenLoopPlant):
    """Emulator of a open-loop RL quadrature plant."""

    REGISTER_SET_POINT = "CL_VOL_Q_SET_POINT"
    REGISTER_COMMAND = "CL_VOL_Q_CMD"
    REGISTER_VALUE = "CL_CUR_Q_VALUE"

    def __init__(self, drive: "VirtualDrive") -> None:
        self.l_henry = L_VALUE
        self.r_ohm = R_VALUE
        super().__init__(drive, [1], [self.l_henry, self.r_ohm])


class PlantClosedLoopRLQuadrature(BaseClosedLoopPlant):
    """Emulator of a closed-loop RL quadrature plant."""

    REGISTER_SET_POINT = "CL_CUR_Q_SET_POINT"
    REGISTER_COMMAND = "CL_CUR_Q_REF_VALUE"
    REGISTER_VALUE = "CL_CUR_Q_VALUE"

    KI_REG = "CL_CUR_Q_KI"
    KP_REG = "CL_CUR_Q_KP"


class PlantOpenLoopJB(BaseOpenLoopPlant):
    """Emulator of a open-loop JB plant."""

    REGISTER_SET_POINT = "CL_CUR_Q_SET_POINT"
    REGISTER_COMMAND = "CL_CUR_Q_VALUE"
    REGISTER_VALUE = "CL_VEL_FBK_VALUE"

    def __init__(self, drive: "VirtualDrive") -> None:
        self.j_value = J_VALUE
        self.b_value = B_VALUE
        super().__init__(drive, [1], [self.j_value, self.b_value])


class PlantClosedLoopJB(BaseClosedLoopPlant):
    """Emulator of a closed-loop JB plant."""

    REGISTER_SET_POINT = "CL_VEL_SET_POINT_VALUE"
    REGISTER_COMMAND = "CL_VEL_REF_VALUE"
    REGISTER_VALUE = "CL_VEL_FBK_VALUE"

    KI_REG = "CL_VEL_PID_KI"
    KP_REG = "CL_VEL_PID_KP"

    TUNING_OPERATION_MODE = OperationMode.PROFILE_VELOCITY
    OL_PLANT_INPUT = "CL_CUR_Q_CMD_VALUE"


class PlantOpenLoopPosition(BaseOpenLoopPlant):
    """Emulator of a open-loop position plant."""

    REGISTER_SET_POINT = "CL_VEL_SET_POINT_VALUE"
    REGISTER_COMMAND = "CL_VEL_FBK_VALUE"
    REGISTER_VALUE = "CL_POS_FBK_VALUE"

    def __init__(self, drive: "VirtualDrive") -> None:
        pos_to_vel_ratio = int(drive.get_value_by_id(1, "PROF_POS_VEL_RATIO"))
        resolution = int(drive.get_value_by_id(1, "FBK_DIGENC1_RESOLUTION"))
        super().__init__(drive, [pos_to_vel_ratio * resolution], [1, 0])


class PlantClosedLoopPosition(BaseClosedLoopPlant):
    """Emulator of a closed-loop position plant."""

    REGISTER_SET_POINT = "CL_POS_SET_POINT_VALUE"
    REGISTER_COMMAND = "CL_POS_REF_VALUE"
    REGISTER_VALUE = "CL_POS_FBK_VALUE"

    KI_REG = "CL_POS_PID_KI"
    KP_REG = "CL_POS_PID_KP"

    TUNING_OPERATION_MODE = OperationMode.PROFILE_POSITION_S_CURVE
    OL_PLANT_INPUT = "CL_VEL_CMD_VALUE"


class PlantOpenLoopVoltageToVelocity(BaseOpenLoopPlant):
    """Emulator of a open-loop voltage-to-velocity plant."""

    REGISTER_SET_POINT = "CL_VOL_Q_SET_POINT"
    REGISTER_COMMAND = "CL_VOL_Q_REF_VALUE"
    REGISTER_VALUE = "CL_VEL_FBK_VALUE"

    def __init__(
        self,
        drive: "VirtualDrive",
    ) -> None:
        self.gain = TORQUE_CONSTANT / (R_VALUE * B_VALUE)
        super().__init__(drive, [self.gain], [1])
        self.plant = signal.TransferFunction([self.gain], [1], dt=1 / self.monitoring_frequency)


class PlantOpenLoopVoltageToCurrentA(PlantOpenLoopRL):
    """Emulator of a open-loop voltage-to-current (A) plant."""

    REGISTER_SET_POINT = "CL_VOL_D_SET_POINT"
    REGISTER_COMMAND = "CL_VOL_D_REF_VALUE"
    REGISTER_VALUE = "FBK_CUR_A_VALUE"


class PlantOpenLoopVoltageToCurrentB(PlantOpenLoopVoltageToCurrentA):
    """Emulator of a open-loop voltage-to-current (B) plant."""

    REGISTER_VALUE = "FBK_CUR_B_VALUE"


class PlantOpenLoopVoltageToCurrentC(PlantOpenLoopVoltageToCurrentA):
    """Emulator of a open-loop voltage-to-current (C) plant."""

    REGISTER_VALUE = "FBK_CUR_C_VALUE"


class VirtualPhasing:
    """Mock of the phasing calibration.

    Args:
        drive: Instance of the VirtualDrive.
    """

    PHASING_MODE_REGISTER = "COMMU_PHASING_MODE"
    PHASING_CURRENT_REGISTER = "COMMU_PHASING_MAX_CURRENT"
    PHASING_TIMEOUT_REGISTER = "COMMU_PHASING_TIMEOUT"
    PHASING_ACCURACY_REGISTER = "COMMU_PHASING_ACCURACY"

    REFERENCE_ANGLE_VALUE_REGISTER = "COMMU_ANGLE_REF_VALUE"
    COMMUTATION_ANGLE_VALUE_REGISTER = "COMMU_ANGLE_VALUE"
    REFERENCE_ANGLE_OFFSET_REGISTER = "COMMU_ANGLE_REF_OFFSET"
    COMMUTATION_ANGLE_OFFSET_REGISTER = "COMMU_ANGLE_OFFSET"

    REFERENCE_FEEDBACK_SENSOR = "COMMU_ANGLE_REF_SENSOR"

    CURRENT_CMD_VALUE_REGISTER = "CL_CUR_CMD_VALUE"
    CURRENT_ACTUAL_VALUE_REGISTER = "FBK_CUR_MODULE_VALUE"

    SETUP_TIME = 1.0
    INITIAL_ANGLE = 180.0
    INITIAL_ANGLE_HALLS = 240.0
    SAMPLING_RATE = 100

    def __init__(self, drive: "VirtualDrive") -> None:
        self.drive = drive
        self.start_time = 0.0
        self.phased = False
        self.drive.reg_signals[self.REFERENCE_ANGLE_VALUE_REGISTER] = (
            np.ones(1) * self.reference_angle_offset
        )
        self.drive.reg_time[self.REFERENCE_ANGLE_VALUE_REGISTER] = np.zeros(1)
        self.drive.reg_signals[self.COMMUTATION_ANGLE_VALUE_REGISTER] = (
            np.ones(1) * self.commutation_angle_offset
        )
        self.drive.reg_time[self.COMMUTATION_ANGLE_VALUE_REGISTER] = np.zeros(1)
        self.drive.reg_signals[self.CURRENT_CMD_VALUE_REGISTER] = np.zeros(1)
        self.drive.reg_time[self.CURRENT_CMD_VALUE_REGISTER] = np.zeros(1)
        self.drive.reg_signals[self.CURRENT_ACTUAL_VALUE_REGISTER] = np.zeros(1)
        self.drive.reg_time[self.CURRENT_ACTUAL_VALUE_REGISTER] = np.zeros(1)
        self.drive.reg_noise_amplitude[self.CURRENT_ACTUAL_VALUE_REGISTER] = 0.01

    def enable(self) -> None:
        """Enable phasing and create signals."""
        if self.phasing_mode != PhasingMode.FORCED or self.phasing_bit:
            return
        self.start_time = time.time()
        period = self.SETUP_TIME + self.phasing_steps * self.phasing_timeout
        n_samples = int(self.SAMPLING_RATE * period)
        time_vector = self.start_time - self.SETUP_TIME + np.linspace(0, period, n_samples)

        setup_ix = int(self.SETUP_TIME * self.SAMPLING_RATE)
        step_samples = int(self.phasing_timeout * self.SAMPLING_RATE)

        commutation_signal = np.zeros(n_samples)
        commutation_signal[:setup_ix] = self.commutation_angle_offset
        commutation_signal[setup_ix:] = 0.0

        reference_signal = np.zeros(n_samples)
        reference_signal[:setup_ix] = self.commutation_angle_offset
        initial_angle_rev = self.initial_angle / 360
        actual_angle_rev = initial_angle_rev
        for step in range(self.phasing_steps):
            step_ix = step * step_samples + setup_ix
            reference_signal[step_ix : step_ix + step_samples] = actual_angle_rev
            actual_angle_rev += (1 - actual_angle_rev) / 2

        current_signal = np.zeros(n_samples)
        current_signal[:setup_ix] = 0.0
        current_signal[setup_ix : step_samples + setup_ix] = np.linspace(
            0, self.phasing_current, step_samples
        )
        current_signal[step_samples + setup_ix :] = self.phasing_current

        self.drive.reg_signals[self.REFERENCE_ANGLE_VALUE_REGISTER] = reference_signal
        self.drive.reg_time[self.REFERENCE_ANGLE_VALUE_REGISTER] = time_vector
        self.drive.reg_signals[self.COMMUTATION_ANGLE_VALUE_REGISTER] = commutation_signal
        self.drive.reg_time[self.COMMUTATION_ANGLE_VALUE_REGISTER] = time_vector
        self.drive.reg_signals[self.CURRENT_CMD_VALUE_REGISTER] = current_signal
        self.drive.reg_time[self.CURRENT_CMD_VALUE_REGISTER] = time_vector
        self.drive.reg_signals[self.CURRENT_ACTUAL_VALUE_REGISTER] = current_signal
        self.drive.reg_time[self.CURRENT_ACTUAL_VALUE_REGISTER] = time_vector

        self.phased = True
        Timer(
            self.phasing_steps * self.phasing_timeout,
            self.set_phasing_bit,
            args=(reference_signal[-1],),
        ).start()

    @property
    def phasing_steps(self) -> int:
        """Phasing steps."""
        delta = 3 * self.phasing_accuracy / 1000

        actual_angle = self.initial_angle
        num_of_steps = 1
        while actual_angle > delta:
            actual_angle = actual_angle / 2
            num_of_steps += 1
        return num_of_steps

    def set_phasing_bit(self, new_offset: Optional[float] = None) -> None:
        """Set the phasing bit.

        Args:
            new_offset: New offset to be set.
        """
        if not self.phased:
            return
        new_status_word = self.drive.status_word | STATUS_WORD_COMMUTATION_FEEDBACK_ALIGNED_BIT
        self.drive.status_word = new_status_word
        if new_offset:
            self.reference_angle_offset = new_offset
            self.commutation_angle_offset = new_offset

    def clear_phasing_bit(self) -> None:
        """Clear phasing bit."""
        new_status_word = self.drive.status_word & ~STATUS_WORD_COMMUTATION_FEEDBACK_ALIGNED_BIT
        self.drive.status_word = new_status_word

    @property
    def phasing_bit(self) -> bool:
        """Phasing bit."""
        phasing_bit = self.drive.status_word & STATUS_WORD_COMMUTATION_FEEDBACK_ALIGNED_BIT
        return bool(phasing_bit)

    @property
    def phasing_mode(self) -> int:
        """Phasing mode."""
        return int(self.drive.get_value_by_id(1, self.PHASING_MODE_REGISTER))

    @property
    def phasing_current(self) -> int:
        """Phasing max current."""
        return int(self.drive.get_value_by_id(1, self.PHASING_CURRENT_REGISTER))

    @property
    def phasing_timeout(self) -> float:
        """Phasing timeout."""
        return float(self.drive.get_value_by_id(1, self.PHASING_TIMEOUT_REGISTER)) / 1000

    @property
    def phasing_accuracy(self) -> int:
        """Phasing accuracy."""
        return int(self.drive.get_value_by_id(1, self.PHASING_ACCURACY_REGISTER))

    @property
    def reference_feedback(self) -> int:
        """Reference feedback sensor."""
        return int(self.drive.get_value_by_id(1, self.REFERENCE_FEEDBACK_SENSOR))

    @property
    def reference_angle_offset(self) -> float:
        """Reference angle offset."""
        return float(self.drive.get_value_by_id(1, self.REFERENCE_ANGLE_OFFSET_REGISTER))

    @reference_angle_offset.setter
    def reference_angle_offset(self, value: float) -> None:
        self.drive.set_value_by_id(1, self.REFERENCE_ANGLE_OFFSET_REGISTER, value)

    @property
    def commutation_angle_offset(self) -> float:
        """Commutation angle offset."""
        return float(self.drive.get_value_by_id(1, self.COMMUTATION_ANGLE_OFFSET_REGISTER))

    @commutation_angle_offset.setter
    def commutation_angle_offset(self, value: float) -> None:
        self.drive.set_value_by_id(1, self.COMMUTATION_ANGLE_OFFSET_REGISTER, value)

    @property
    def initial_angle(self) -> float:
        """Initial angle."""
        if self.reference_feedback == SensorType.HALLS:
            return self.INITIAL_ANGLE_HALLS
        else:
            return self.INITIAL_ANGLE


class VirtualInternalGenerator:
    """Emulate the virtual generator with the only purpose of mocking the feedback tests.

    Args:
        drive: Instance of VirtualDrive.
    """

    MODE_REGISTER = "FBK_GEN_MODE"
    FREQUENCY_REGISTER = "FBK_GEN_FREQ"
    GAIN_REGISTER = "FBK_GEN_GAIN"
    OFFSET_REGISTER = "FBK_GEN_OFFSET"
    CYCLE_NUMBER_REGISTER = "FBK_GEN_CYCLES"
    REARM_REGISTER = "FBK_GEN_REARM"

    ACTUAL_POSITION_REGISTER = "CL_POS_FBK_VALUE"
    POLE_PAIRS_REGISTER = "MOT_PAIR_POLES"
    DIG_HALL_POLE_PAIRS_REGISTER = "FBK_DIGHALL_PAIRPOLES"
    ABS1_ST_BITS_REGISTER = "FBK_BISS1_SSI1_POS_ST_BITS"
    ABS2_ST_BITS_REGISTER = "FBK_BISS2_POS_ST_BITS"
    COMMUTATION_FEEDBACK_REGISTER = "COMMU_ANGLE_SENSOR"
    POSITION_FEEDBACK_REGISTER = "CL_POS_FBK_SENSOR"

    HALL_VALUES = [1, 3, 2, 6, 4, 5]

    ENCODER_REGISTERS = {
        SensorType.QEI: "FBK_DIGENC1_VALUE",
        SensorType.QEI2: "FBK_DIGENC2_VALUE",
        SensorType.HALLS: "FBK_DIGHALL_VALUE",
        SensorType.ABS1: "FBK_BISS1_SSI1_POS_VALUE",
        SensorType.BISSC2: "FBK_BISS2_POS_VALUE",
    }

    def __init__(self, drive: "VirtualDrive") -> None:
        self.drive = drive
        self.start_time = 0.0

    def enable(self) -> None:
        """Enable internal generator and generate the encoder and position signals."""
        self.start_time = time.time()
        if (
            self.commutation_feedback != SensorType.INTGEN
            or self.generator_mode != GeneratorMode.SAW_TOOTH
            or self.position_encoder not in self.ENCODER_REGISTERS
        ):
            return
        if self.position_encoder == SensorType.HALLS:
            pole_pairs = self.drive.get_value_by_id(1, self.POLE_PAIRS_REGISTER)
            self.drive.set_value_by_id(1, self.DIG_HALL_POLE_PAIRS_REGISTER, pole_pairs)

        period = 1 / self.frequency
        n_samples = int(VirtualMonitoring.FREQUENCY * period * self.cycles)
        time_vector = self.start_time + np.linspace(0, period * self.cycles, n_samples)

        signal_period: NDArray[np.float_] = self.offset + np.linspace(
            0, self.gain, int(n_samples / self.cycles), dtype=np.float_
        )
        pos_signal = np.tile(signal_period, self.cycles)
        initial_value = self.drive.get_value_by_id(1, self.ACTUAL_POSITION_REGISTER)
        pos_signal = initial_value + (pos_signal - pos_signal[0]) * self.encoder_resolution
        pos_signal = pos_signal.astype(int)

        if self.position_encoder == SensorType.HALLS:
            encoder_signal = self.__create_halls_encoder_signal(pos_signal)
        elif self.position_encoder in [SensorType.ABS1, SensorType.BISSC2]:
            encoder_signal = self.__create_abs_encoder_signal(pos_signal)
        else:
            encoder_signal = pos_signal.copy()

        self.drive.reg_signals[self.encoder_register] = encoder_signal
        self.drive.reg_time[self.encoder_register] = time_vector
        self.drive.reg_signals[self.ACTUAL_POSITION_REGISTER] = pos_signal
        self.drive.reg_time[self.ACTUAL_POSITION_REGISTER] = time_vector

    def __create_halls_encoder_signal(self, pos_signal: NDArray[np.float_]) -> NDArray[np.float_]:
        """Create the halls encoder signal by discretizing the pos_signal using the hall values.

        Args:
            pos_signal: Position signal.

        Returns:
            Encoder signal.

        """
        encoder_signal = pos_signal.copy()
        encoder_signal[0] = self.drive.get_value_by_id(1, self.encoder_register)
        hall_value_ix = self.HALL_VALUES.index(encoder_signal[0])
        for sample_ix in range(1, len(encoder_signal)):
            hall_change = 0
            if (pos_signal[sample_ix] - pos_signal[sample_ix - 1]) > 0:
                hall_change = 1
            elif (pos_signal[sample_ix] - pos_signal[sample_ix - 1]) < 0:
                hall_change = -1
            else:
                hall_change = 0
            hall_value_ix = (hall_value_ix + hall_change) % len(self.HALL_VALUES)
            encoder_signal[sample_ix] = self.HALL_VALUES[hall_value_ix]
        return encoder_signal

    def __create_abs_encoder_signal(self, pos_signal: NDArray[np.float_]) -> NDArray[np.float_]:
        """Crete the absolute encoder signal by saturating the position signal.

        Args:
            pos_signal: Position signal.

        Returns:
            Encoder signal.
        """
        encoder_signal = pos_signal.copy()
        for sample_ix in range(1, len(encoder_signal)):
            if encoder_signal[sample_ix] > self.encoder_resolution - 1:
                encoder_signal[sample_ix:] = encoder_signal[sample_ix:] - self.encoder_resolution
            elif encoder_signal[sample_ix] < 0:
                encoder_signal[sample_ix:] = encoder_signal[sample_ix:] + self.encoder_resolution
        return encoder_signal

    @property
    def encoder_register(self) -> str:
        """Register of the encoder value."""
        return self.ENCODER_REGISTERS[SensorType(self.position_encoder)]

    @property
    def encoder_resolution(self) -> int:
        """Encoder resolution."""
        if self.position_encoder == SensorType.QEI:
            return INC_ENC1_RESOLUTION
        elif self.position_encoder == SensorType.QEI2:
            return INC_ENC2_RESOLUTION
        elif self.position_encoder == SensorType.HALLS:
            pole_pairs = int(self.drive.get_value_by_id(1, self.DIG_HALL_POLE_PAIRS_REGISTER))
            return pole_pairs * len(self.HALL_VALUES)
        elif self.position_encoder == SensorType.ABS1:
            single_turn_bits = int(self.drive.get_value_by_id(1, self.ABS1_ST_BITS_REGISTER))
            return int(2**single_turn_bits)
        elif self.position_encoder == SensorType.BISSC2:
            single_turn_bits = int(self.drive.get_value_by_id(1, self.ABS2_ST_BITS_REGISTER))
            return int(2**single_turn_bits)
        else:
            return 1

    @property
    def position_encoder(self) -> int:
        """Position encoder."""
        return int(self.drive.get_value_by_id(1, self.POSITION_FEEDBACK_REGISTER))

    @property
    def generator_mode(self) -> int:
        """Generator mode."""
        return int(self.drive.get_value_by_id(1, self.MODE_REGISTER))

    @property
    def commutation_feedback(self) -> int:
        """Commutation feedback sensor."""
        return int(self.drive.get_value_by_id(1, self.COMMUTATION_FEEDBACK_REGISTER))

    @property
    def frequency(self) -> float:
        """Internal generator frequency."""
        return float(self.drive.get_value_by_id(1, self.FREQUENCY_REGISTER))

    @property
    def offset(self) -> float:
        """Internal generator offset."""
        return float(self.drive.get_value_by_id(1, self.OFFSET_REGISTER))

    @property
    def gain(self) -> float:
        """Internal generator gain."""
        return float(self.drive.get_value_by_id(1, self.GAIN_REGISTER))

    @property
    def cycles(self) -> int:
        """Internal generator cycles."""
        return int(self.drive.get_value_by_id(1, self.CYCLE_NUMBER_REGISTER))


class VirtualMonDistBase:
    """Base class to implement VirtualMonitoring and VirtualDisturbance.

    Args:
        drive: Virtual drive instance.

    """

    FREQUENCY = 20000
    FREQ_DIVIDER_REG: str
    BUFFER_SIZE_REG: str
    NUMBER_MAP_REGS: str
    MAP_REG_CFG: str
    BYTES_PER_BLOCK_REG: str
    AVAILABLE_BYTES_REG: str
    DATA_REG: str
    STATUS_REGISTER: str

    def __init__(self, drive: "VirtualDrive") -> None:
        self.drive = drive
        self.enabled = False
        self.disable()
        self.number_mapped_registers = 0
        self.channels_data: Dict[int, List[Union[int, float]]] = {}
        self.channels_dtype: Dict[int, REG_DTYPE] = {}
        self.channels_address: Dict[int, int] = {}
        self.channels_subnode: Dict[int, int] = {}
        self.channels_size: Dict[int, int] = {}
        self.channels_signal: Dict[int, List[Union[int, float]]] = {}

    def enable(self) -> None:
        """Enable Monitoring/Disturbance."""
        self.enabled = True
        self.drive.set_value_by_id(0, self.STATUS_REGISTER, 1)

    def disable(self) -> None:
        """Disable Monitoring/Disturbance."""
        self.enabled = False
        self.drive.set_value_by_id(0, self.STATUS_REGISTER, 0)
        self.remove_data()

    def remove_data(self) -> None:
        """Remove Monitoring/Disturbance data."""
        self.channels_data = {}
        self.drive.set_value_by_id(0, self.DATA_REG, bytes(1))

    @property
    def divider(self) -> int:
        """Frequency divider."""
        value = self.drive.get_value_by_id(0, self.FREQ_DIVIDER_REG)
        if isinstance(value, int):
            return value
        else:
            return 0

    @property
    def buffer_size(self) -> int:
        """Monitoring buffer size."""
        value = self.drive.get_value_by_id(0, self.BUFFER_SIZE_REG)
        if isinstance(value, int):
            return value
        else:
            return 0

    @property
    def buffer_time(self) -> float:
        """Monitoring buffer size in seconds."""
        return self.buffer_size * self.divider / self.FREQUENCY

    @property
    def bytes_per_block(self) -> int:
        """Monitoring bytes per sample."""
        value = self.drive.get_value_by_id(0, self.BYTES_PER_BLOCK_REG)
        if isinstance(value, int):
            return value
        else:
            return 0

    @bytes_per_block.setter
    def bytes_per_block(self, n_bytes: int) -> None:
        self.drive.set_value_by_id(0, self.BYTES_PER_BLOCK_REG, n_bytes)

    @property
    def number_mapped_registers(self) -> int:
        """Number of mapped registers."""
        value = self.drive.get_value_by_id(0, self.NUMBER_MAP_REGS)
        if isinstance(value, int):
            return value
        else:
            return 0

    @number_mapped_registers.setter
    def number_mapped_registers(self, value: int) -> None:
        self.drive.set_value_by_id(0, self.NUMBER_MAP_REGS, value)

    @property
    def available_bytes(self) -> int:
        """Actual number of monitoring bytes."""
        value = self.drive.get_value_by_id(0, self.AVAILABLE_BYTES_REG)
        if isinstance(value, int):
            return value
        else:
            return 0

    @available_bytes.setter
    def available_bytes(self, n_bytes: int) -> None:
        self.drive.set_value_by_id(0, self.AVAILABLE_BYTES_REG, n_bytes)

    def get_mapped_register(self, channel: int) -> Tuple[int, int, int, int]:
        """Decodes the register with the information of a mapped register.

        Args:
            channel: Channel of the register to be decoded.

        Returns:
            Register subnode
            Register address
            Register dtype index
            Channel size

        """
        register_id = self.MAP_REG_CFG.format(channel)
        data = self.drive.get_value_by_id(0, register_id)
        if not isinstance(data, int):
            raise ValueError("Wrong register type")
        data_h = data >> 16
        data_l = data & 0x0000FFFF
        subnode = data_h >> 12
        address = data_h & 0x0FFF
        dtype = data_l >> 8
        size = data_l & 0x00FF
        return subnode, address, dtype, size

    def map_registers(self) -> None:
        """Creates the channels attribute based on mapped registers."""
        self.bytes_per_block = 0
        empty_list: List[float] = []
        for channel in range(self.number_mapped_registers):
            subnode, address, dtype, size = self.get_mapped_register(channel)
            self.channels_data[channel] = empty_list.copy()
            self.channels_dtype[channel] = REG_DTYPE(dtype)
            self.channels_address[channel] = address
            self.channels_subnode[channel] = subnode
            self.channels_size[channel] = size
            self.channels_signal[channel] = empty_list.copy()
            self.bytes_per_block += size

    def get_channel_index(self, reg_id: str) -> Optional[int]:
        """Return the channel index of a mapped register.

        If the register is not mapped returns None.

        Args:
            reg_id: Register ID.

        Returns:
            Channel index. None if the register is not mapped.
        """
        channel_index = None
        for channel in range(self.number_mapped_registers):
            subnode, address, _, _ = self.get_mapped_register(channel)
            mapped_register = self.drive.get_register(subnode, address)
            if mapped_register.identifier == reg_id:
                channel_index = channel
                break

        return channel_index


class VirtualMonitoring(VirtualMonDistBase):
    """Emulates monitoring at the VirtualDrive.

    Args:
        drive: Virtual drive instance.

    """

    FREQ_DIVIDER_REG = "MON_DIST_FREQ_DIV"
    BUFFER_SIZE_REG = "MON_CFG_WINDOW_SAMP"
    NUMBER_MAP_REGS = "MON_CFG_TOTAL_MAP"
    MAP_REG_CFG = "MON_CFG_REG{}_MAP"
    BYTES_PER_BLOCK_REG = "MON_CFG_BYTES_PER_BLOCK"
    DATA_REG = "MONITORING_DATA"
    TRIGGER_TYPE_REG = "MON_CFG_SOC_TYPE"
    AVAILABLE_BYTES_REG = "MON_CFG_BYTES_VALUE"
    STATUS_REGISTER = "MON_DIST_STATUS"

    def __init__(self, drive: "VirtualDrive") -> None:
        self.start_time = 0.0
        super().__init__(drive)

    def enable(self) -> None:
        super().map_registers()
        super().enable()

        # Set monitoring end and frame available
        self.drive.set_value_by_id(0, self.STATUS_REGISTER, 0x10 | (0x8 + 1))
        self.update_data()

    def update_data(self) -> None:
        self.__create_signals()

        # Store data
        sampling_rate = self.FREQUENCY / self.divider
        elapsed_time = time.time() - self.start_time
        elapsed_samples = int(elapsed_time * sampling_rate)
        n_samples = min(elapsed_samples, self.buffer_size)
        for channel in range(self.number_mapped_registers):
            self.channels_data[channel] = self.channels_signal[channel][:n_samples]
        self.available_bytes = n_samples * self.bytes_per_block
        self._store_data_bytes()
        self.start_time = 0.0

    def trigger(self) -> None:
        """Triggers monitoring."""
        if self.enabled:
            self.start_time = time.time()

    def rearm(self) -> None:
        """Rearm monitoring."""
        self.map_registers()
        self.disable()
        self.enable()

    def __create_signals(self) -> None:
        """Creates emulated monitoring signals."""
        for channel in range(self.number_mapped_registers):
            subnode = self.channels_subnode[channel]
            address = self.channels_address[channel]
            reg_id = self.drive.address_to_id(subnode, address)
            if len(self.drive.reg_signals[reg_id]) > 0:
                self.channels_signal[channel] = self.drive.reg_signals[reg_id]
                if self.divider > 1:
                    indexes = np.arange(
                        0, self.buffer_size * self.divider - 1, self.divider, dtype=int
                    )
                    self.channels_signal[channel] = self.channels_signal[channel][indexes]
                if self.drive.reg_noise_amplitude[reg_id] > 0:
                    noise = self.drive.reg_noise_amplitude[reg_id] * np.random.normal(
                        size=len(self.channels_signal[channel])
                    )
                    self.channels_signal[channel] = self.channels_signal[channel] + noise
                continue

            start_value = address + subnode
            if self.channels_dtype[channel] == REG_DTYPE.FLOAT:
                signal = [
                    float(start_value + i)
                    for i in range(0, self.buffer_size * self.divider, self.divider)
                ]
            else:
                signal = [
                    start_value + i for i in range(0, self.buffer_size * self.divider, self.divider)
                ]
            self.channels_signal[channel] = signal

    def _store_data_bytes(self) -> None:
        """Convert signals into a bytes and store it at MON_DATA register."""
        byte_array = bytes()
        n_samples = len(self.channels_data[0])
        for sample in range(n_samples):
            for channel in range(self.number_mapped_registers):
                value = self.channels_data[channel][sample]
                size = self.channels_size[channel]
                if self.channels_dtype[channel] != REG_DTYPE.FLOAT:
                    value = int(value)
                sample_bytes = convert_dtype_to_bytes(value, self.channels_dtype[channel])
                if len(sample_bytes) < size:
                    sample_bytes += b"0" * (size - len(sample_bytes))
                byte_array += sample_bytes
        self.drive.set_value_by_id(0, "MONITORING_DATA", byte_array)

    @property
    def trigger_type(self) -> int:
        """Trigger type Auto(0), Force (1) or Rising or Failing (2)."""
        value = self.drive.get_value_by_id(0, self.TRIGGER_TYPE_REG)
        if isinstance(value, int):
            return value
        else:
            return 0


class VirtualDisturbance(VirtualMonDistBase):
    """Emulates disturbance at the VirtualDrive.

    Args:
        drive: Virtual drive instance.

    """

    FREQ_DIVIDER_REG = "DIST_FREQ_DIV"
    BUFFER_SIZE_REG = "DIST_CFG_SAMPLES"
    NUMBER_MAP_REGS = "DIST_CFG_MAP_REGS"
    MAP_REG_CFG = "DIST_CFG_REG{}_MAP"
    BYTES_PER_BLOCK_REG = "DIST_CFG_BYTES_PER_BLOCK"
    AVAILABLE_BYTES_REG = "DIST_CFG_BYTES"
    DATA_REG = "DISTURBANCE_DATA"
    STATUS_REGISTER = "DIST_STATUS"

    def __init__(self, drive: "VirtualDrive") -> None:
        self.start_time = 0.0
        self.received_bytes = bytes()
        super().__init__(drive)

    def enable(self) -> None:
        self.start_time = time.time()
        for channel in range(self.number_mapped_registers):
            subnode = self.channels_subnode[channel]
            address = self.channels_address[channel]
            reg = self.drive.get_register(subnode, address)
            self.drive.reg_time[str(reg.identifier)] += self.start_time
        super().enable()

    def disable(self) -> None:
        self.received_bytes = bytes()
        return super().disable()

    def append_data(self, data: bytes) -> None:
        """Append received disturbance data until the buffer is full.

        Args:
            data: Received data.
        """
        if len(self.channels_data) == 0:
            super().map_registers()
        self.received_bytes += data
        self.available_bytes = len(self.received_bytes)
        if self.available_bytes == self.buffer_size_bytes:
            self.write_data()

    def write_data(self) -> None:
        """Convert received data and store it at the channels attribute."""
        n_samples = self.buffer_size
        buffer = self.received_bytes
        for channel in range(self.number_mapped_registers):
            for _ in range(n_samples):
                size = self.channels_size[channel]
                dtype = self.channels_dtype[channel]
                bytes = buffer[:size]
                value = convert_bytes_to_dtype(bytes, dtype)
                if isinstance(value, (int, float)):
                    self.channels_data[channel].append(value)
                buffer = buffer[size:]

            dist_signal = np.array(self.channels_data[channel])
            sampling_rate = self.FREQUENCY / self.divider
            total_time = len(dist_signal) / sampling_rate
            time_vector = np.arange(0.0, total_time, 1 / sampling_rate)

            if self.divider > 1:
                time_vector_resampled = np.arange(0.0, total_time, 1 / self.FREQUENCY)
                dist_signal = np.interp(time_vector_resampled, time_vector, dist_signal)
                time_vector = time_vector_resampled

            subnode = self.channels_subnode[channel]
            address = self.channels_address[channel]
            reg_id = str(self.drive.get_register(subnode, address).identifier)

            self.drive.reg_time[reg_id] = time_vector
            self.drive.reg_signals[reg_id] = dist_signal

    @property
    def buffer_size_bytes(self) -> int:
        """Buffer size in bytes."""
        buffer_size_bytes = 0
        n_samples = self.buffer_size
        for channel in range(self.number_mapped_registers):
            buffer_size_bytes += self.channels_size[channel] * n_samples
        return buffer_size_bytes


class VirtualDrive(Thread):
    """Emulates a drive by creating a UDP server that sends and receives MCB messages.

    Args:
        port: Server port number.
        dictionary_path: Path to the dictionary. If None, the default dictionary is used.

    """

    IP_ADDRESS = "127.0.0.1"

    ACK_CMD = 3
    WRITE_CMD = 2
    READ_CMD = 1

    PATH_CONFIGURATION_RELATIVE = "./resources/virtual_drive.xcf"
    PATH_DICTIONARY_RELATIVE = "./resources/virtual_drive.xdf"

    def __init__(self, port: int, dictionary_path: Optional[str] = None) -> None:
        super(VirtualDrive, self).__init__()
        self.ip = self.IP_ADDRESS
        self.port = port
        default_dictionary = os.path.join(
            pathlib.Path(__file__).parent.resolve(), self.PATH_DICTIONARY_RELATIVE
        )
        self.dictionary_path = dictionary_path or default_dictionary
        self.__stop = False
        self.device_info = None
        self.__logger: List[Dict[str, Union[float, bytes, str, Tuple[str, int]]]] = []
        self.__reg_address_to_id: Dict[int, Dict[int, str]] = {}
        self.__dictionary = VirtualDictionary(self.dictionary_path)
        self.reg_signals: Dict[str, NDArray[np.float_]] = {}
        self.reg_time: Dict[str, NDArray[np.float_]] = {}
        self.reg_noise_amplitude: Dict[str, float] = {}

        self._init_registers()
        self._update_registers()
        self._init_register_signals()
        self.__set_motor_ready_to_switch_on()

        self._monitoring: Optional[VirtualMonitoring] = None
        self._disturbance: Optional[VirtualDisturbance] = None
        if self.__register_exists(0, VirtualMonitoring.STATUS_REGISTER):
            self._monitoring = VirtualMonitoring(self)
            self._disturbance = VirtualDisturbance(self)

        self._plant_open_loop_rl_d = PlantOpenLoopRL(self)
        self._plant_closed_loop_rl_d = PlantClosedLoopRL(self, self._plant_open_loop_rl_d.plant)
        self._plant_open_loop_rl_q = PlantOpenLoopRLQuadrature(self)
        self._plant_closed_loop_rl_q = PlantClosedLoopRLQuadrature(
            self, self._plant_open_loop_rl_q.plant
        )
        self._plant_open_loop_jb = PlantOpenLoopJB(self)
        self._plant_closed_loop_jb = PlantClosedLoopJB(self, self._plant_open_loop_jb.plant)
        self._plant_open_loop_position = PlantOpenLoopPosition(self)
        self._plant_closed_loop_position = PlantClosedLoopPosition(
            self, self._plant_open_loop_position.plant
        )
        self._plant_open_loop_vol_to_vel = PlantOpenLoopVoltageToVelocity(self)
        self._plant_open_loop_vol_to_curr_a = PlantOpenLoopVoltageToCurrentA(self)
        self._plant_open_loop_vol_to_curr_b = PlantOpenLoopVoltageToCurrentB(self)
        self._plant_open_loop_vol_to_curr_c = PlantOpenLoopVoltageToCurrentC(self)

        self.phasing = VirtualPhasing(self)
        self.internal_generator = VirtualInternalGenerator(self)

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if platform.system() != "Windows":
            # On Linux, the SO_REUSEADDR should be set to avoid keeping the socket opened.
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def run(self) -> None:
        """Open socket, listen and decode messages."""
        server_address = (self.ip, self.port)
        self.socket.bind(server_address)
        self.socket.settimeout(2)
        while not self.__stop:
            try:
                frame, add = self.socket.recvfrom(ETH_BUF_SIZE)
            except:
                continue
            reg_add, subnode, cmd, data = MCB.read_mcb_frame(frame)
            self.__log(add, frame, MSG_TYPE.RECEIVED)
            register = self.get_register(subnode, reg_add)
            if cmd == self.WRITE_CMD:
                response = self.__get_response_to_write_command(register, data)
            elif cmd == self.READ_CMD:
                response = self.__get_response_to_read_command(register, data)
            else:
                continue
            self.__send(response, add)

        time.sleep(0.1)

    def __get_response_to_write_command(self, register: EthernetRegister, data: bytes) -> bytes:
        """Return the response to a WRITE command.

        Args:
            register: Register instance.
            data: Received data frame.

        Returns:
            bytes: Response to be sent.
        """
        response = MCB.build_mcb_frame(self.ACK_CMD, register.subnode, register.address, data[:8])
        if register.access in [REG_ACCESS.RW, REG_ACCESS.WO]:
            value = convert_bytes_to_dtype(data, register.dtype)
            self.__decode_msg(register.address, register.subnode, data)
            self.set_value_by_id(register.subnode, str(register.identifier), value)

        return response

    def __get_response_to_read_command(self, register: EthernetRegister, data: bytes) -> bytes:
        """Return the response to a READ command.

        Args:
            register: Register instance.
            data: Received data frame.

        Returns:
            bytes: Response to be sent.
        """
        value = self.get_value_by_id(register.subnode, str(register.identifier))
        if (
            register.address == self.id_to_address(0, "MONITORING_DATA")
            and isinstance(value, bytes)
            and self._monitoring
        ):
            self._monitoring.update_data()
            response = self._response_monitoring_data(value)
        else:
            data = convert_dtype_to_bytes(value, register.dtype)
            response = MCB.build_mcb_frame(self.ACK_CMD, register.subnode, register.address, data)

        return response

    def stop(self) -> None:
        """Stop socket."""
        if self.socket is not None:
            self.socket.close()
        self.__stop = True
        if self._monitoring:
            self._monitoring.disable()

    def _init_registers(self) -> None:
        """Initialize the registers using the configuration file."""
        configuration_file = os.path.join(
            pathlib.Path(__file__).parent.resolve(), self.PATH_CONFIGURATION_RELATIVE
        )
        _, registers = EthernetServo._read_configuration_file(configuration_file)
        cast_data = {"float": float, "str": str}
        for element in registers:
            subnode = int(element.attrib["subnode"])
            reg_dtype = element.attrib["dtype"]
            reg_data = element.attrib["storage"]
            if not self.__register_exists(subnode, element.attrib["id"]):
                continue
            self.set_value_by_id(
                subnode,
                element.attrib["id"],
                cast_data.get(reg_dtype, int)(reg_data),
            )
        value: Union[str, int]
        for subnode in self.__dictionary.subnodes:
            for reg_id, reg in self.__dictionary.registers(subnode).items():
                if reg._storage is not None:
                    continue
                elif reg.enums_count > 0:
                    value = next(iter(reg.enums.values()))
                elif reg.dtype == REG_DTYPE.STR:
                    value = ""
                else:
                    value = 0
                self.set_value_by_id(subnode, reg_id, value)

    def _update_registers(self) -> None:
        """Force storage_valid at each register and add registers that are not in the dictionary."""
        for subnode in self.__dictionary.subnodes:
            self.__reg_address_to_id[subnode] = {}
            for reg_id, reg in self.__dictionary.registers(subnode).items():
                if not isinstance(reg, EthernetRegister):
                    raise ValueError(f"Register {reg_id} is not an EthernetRegister")
                self.__reg_address_to_id[subnode][reg.address] = reg_id
                self.__dictionary.registers(subnode)[reg_id].storage_valid = True

        custom_regs = {
            "MONITORING_DATA": self.__dictionary.registers(0)[EthernetServo.MONITORING_DATA],
            "DISTURBANCE_DATA": self.__dictionary.registers(0)[EthernetServo.DIST_DATA],
        }
        for id, reg in custom_regs.items():
            if not isinstance(reg, EthernetRegister):
                raise ValueError
            register = EthernetRegister(
                reg.address, REG_DTYPE.BYTE_ARRAY_512, reg.access, identifier=id, subnode=reg.subnode
            )
            self.__dictionary._add_register_list(register)
            self.__dictionary.registers(reg.subnode)[id].storage_valid = True

            self.__reg_address_to_id[reg.subnode][reg.address] = id

    def _init_register_signals(self) -> None:
        """Init signals, vector time and noise amplitude for each register."""
        for subnode in self.__dictionary.subnodes:
            for reg_id in self.__dictionary.registers(subnode).keys():
                self.reg_signals[reg_id] = np.array([])
                self.reg_time[reg_id] = np.array([])
                self.reg_noise_amplitude[reg_id] = 0.0

    def __send(self, response: bytes, address: Tuple[str, int]) -> None:
        """Send a message and update log.

        Args:
            response: Message to be sent.
            address: IP address and port.
        """
        self.socket.sendto(response, address)
        self.__log(address, response, MSG_TYPE.SENT)

    def _response_monitoring_data(self, data: bytes) -> bytes:
        """Creates a response for monitoring data.

        Args:
            data: Data to be sent.

        Returns:
            MCB frame.
        """
        if not self._monitoring:
            return bytes(1)
        sent_cmd = self.ACK_CMD
        reg_add = self.id_to_address(0, "MONITORING_DATA")
        limit = min(len(data), MONITORING_BUFFER_SIZE)
        response = MCB.build_mcb_frame(sent_cmd, 0, reg_add, data[:limit])
        data_left = data[limit:]
        self.set_value_by_id(0, "MONITORING_DATA", data_left)
        self._monitoring.available_bytes = len(data_left)
        return response

    def __log(self, ip_port: Tuple[str, int], message: bytes, msg_type: MSG_TYPE) -> None:
        """Updates log.

        Args:
            ip_port: IP address and port.
            message: Received or sent message.
            msg_type: Sent or Received.
        """
        self.__logger.append(
            {
                "timestamp": time.time(),
                "ip_port": ip_port,
                "type": msg_type.value,
                "message": message,
            }
        )

    @property
    def log(self) -> List[Dict[str, Union[float, bytes, str, Tuple[str, int]]]]:
        """Dictionary containing log information."""
        return self.__logger

    def clean_log(self) -> None:
        """Cleans log."""
        self.__logger = []

    def __decode_msg(self, reg_add: int, subnode: int, data: bytes) -> None:
        """Decodes received messages and run specific methods if needed.

        Args:
            reg_add: Register address.
            subnode: Subnode.
            data: Received data.
        """
        register = self.get_register(subnode, reg_add)
        reg_id = register.identifier
        dtype = register.dtype
        value = convert_bytes_to_dtype(data, dtype)
        if reg_id == "MON_DIST_ENABLE" and subnode == 0 and value == 1 and self._monitoring:
            self._monitoring.enable()
        if reg_id == "MON_DIST_ENABLE" and subnode == 0 and value == 0 and self._monitoring:
            self._monitoring.disable()
        if reg_id == "MON_CMD_FORCE_TRIGGER" and subnode == 0 and value == 1 and self._monitoring:
            self._monitoring.trigger()
        if reg_id == "MON_REMOVE_DATA" and subnode == 0 and value == 1 and self._monitoring:
            self._monitoring.remove_data()
        if reg_id == "MON_REARM" and subnode == 0 and value == 1 and self._monitoring:
            self._monitoring.rearm()
            self.__emulate_plants()
        if reg_id == "DIST_ENABLE" and subnode == 0 and value == 1 and self._disturbance:
            self._disturbance.enable()
            self.__emulate_plants()
        if reg_id == "DIST_ENABLE" and subnode == 0 and value == 0 and self._disturbance:
            self._disturbance.disable()
            self.__clean_plant_signals()
        if reg_id == "DIST_REMOVE_DATA" and subnode == 0 and value == 1 and self._disturbance:
            self._disturbance.remove_data()
        if reg_id == "DISTURBANCE_DATA" and subnode == 0 and self._disturbance:
            self._disturbance.append_data(data)
        if reg_id == "DRV_OP_CMD":
            self.operation_mode = int(value)
            self.__clean_plant_signals()
        if reg_id == "DRV_STATE_CONTROL" and subnode == 1 and (int(value) & IL_MC_CW_EO):
            self.__set_motor_enable()
        if reg_id == "DRV_STATE_CONTROL" and subnode == 1 and (value == constants.IL_MC_PDS_CMD_DV):
            self.__set_motor_disable()
        if reg_id == "DRV_STATE_CONTROL" and subnode == 1 and (value == constants.IL_MC_PDS_CMD_SD):
            self.__set_motor_ready_to_switch_on()
        if reg_id == "COMMU_ANGLE_SENSOR" and subnode == 1:
            self.phasing.clear_phasing_bit()
        if (
            reg_id == PlantOpenLoopJB.REGISTER_SET_POINT
            and subnode == 1
            and self.operation_mode in [OperationMode.CURRENT]
        ):
            self._plant_open_loop_jb.jog(float(value))
        if (
            reg_id == PlantClosedLoopRL.REGISTER_SET_POINT
            and subnode == 1
            and self.operation_mode in [OperationMode.CURRENT, OperationMode.VOLTAGE]
        ):
            self._plant_closed_loop_rl_d.jog(float(value))
        if (
            reg_id == PlantClosedLoopRLQuadrature.REGISTER_SET_POINT
            and subnode == 1
            and self.operation_mode in [OperationMode.CURRENT, OperationMode.VOLTAGE]
        ):
            self._plant_closed_loop_rl_q.jog(float(value))
        if (
            reg_id == PlantClosedLoopJB.REGISTER_SET_POINT
            and subnode == 1
            and self.operation_mode in [OperationMode.VELOCITY, OperationMode.PROFILE_VELOCITY]
        ):
            self._plant_closed_loop_jb.jog(float(value))
        if (
            reg_id == PlantClosedLoopPosition.REGISTER_SET_POINT
            and subnode == 1
            and self.operation_mode
            in [
                OperationMode.POSITION,
                OperationMode.PROFILE_POSITION,
                OperationMode.PROFILE_POSITION_S_CURVE,
            ]
        ):
            self._plant_closed_loop_position.jog(int(value))
        if (
            reg_id == PlantOpenLoopVoltageToVelocity.REGISTER_SET_POINT
            and subnode == 1
            and self.operation_mode in [OperationMode.VOLTAGE]
        ):
            self._plant_open_loop_vol_to_vel.jog(float(value))
        if reg_id == VirtualInternalGenerator.REARM_REGISTER and subnode == 1:
            self.internal_generator.enable()

    def address_to_id(self, subnode: int, address: int) -> str:
        """Converts a register address into its ID.

        Args:
            subnode: Subnode.
            address: Register address.

        Returns:
            Register ID.
        """
        return self.__reg_address_to_id[subnode][address]

    def id_to_address(self, subnode: int, uid: str) -> int:
        """Return address of Register with the target subnode and UID

        Args:
            subnode: Subnode.
            uid: Register UID.

        Returns:
            Register address.
        """
        register = self.__dictionary.registers(subnode)[uid]
        if not isinstance(register, EthernetRegister):
            raise ValueError(f"Register {uid} is not an EthernetRegister")
        return register.address

    def get_value_by_id(self, subnode: int, id: str) -> Union[int, float, str, bytes]:
        """Returns a register value by its ID.

        Args:
            subnode: Subnode.
            id: Register ID.

        Returns:
            Register value.
        """
        register = self.__dictionary.registers(subnode)[id]
        value: Union[int, float]
        if len(self.reg_signals[id]) > 0:
            actual_time = time.time()
            if self._disturbance and self._disturbance.enabled:
                time_diff = actual_time - self._disturbance.start_time
                actual_time = self._disturbance.start_time + (
                    time_diff % self._disturbance.buffer_time
                )
            sample_index = np.argmin(np.abs(self.reg_time[id] - actual_time))
            value = self.reg_signals[id][sample_index]
            if self.reg_noise_amplitude[id] > 0:
                value = value + self.reg_noise_amplitude[id] * np.random.uniform()

            if register.dtype != REG_DTYPE.FLOAT:
                value = int(value)
            return value
        storage_value = self.__dictionary.registers(subnode)[id]._storage
        if isinstance(storage_value, (int, float, str, bytes)):
            return storage_value
        else:
            return 0

    def set_value_by_id(self, subnode: int, id: str, value: Union[float, int, str, bytes]) -> None:
        """Set a register value by its ID.

        Args:
            subnode: Subnode.
            id: Register ID.
            value: Value to be set.
        """
        self.__dictionary.registers(subnode)[id].storage = value

    def __register_exists(self, subnode: int, id: str) -> bool:
        """Returns True if the register exists in the dictionary

        Args:
            subnode: Subnode.
            id: Register ID.

        Returns:
            Register value.
        """
        return subnode in self.__dictionary.subnodes and id in self.__dictionary.registers(subnode)

    def get_register(
        self, subnode: int, address: Optional[int] = None, id: Optional[str] = None
    ) -> EthernetRegister:
        """Returns a register by its address or ID.

        Args:
            subnode: Subnode.
            address: Register address. Default to None.
            id: Register ID. Default to None.

        Returns:
            Register instance.

        Raises:
            ValueError: If both address and id are None.
        """
        if address is not None:
            id = self.address_to_id(subnode, address)
        else:
            if id is None:
                raise ValueError("Register address or id should be passed")
        register = self.__dictionary.registers(subnode)[id]
        if not isinstance(register, EthernetRegister):
            raise ValueError(f"Register {id} is not an EthernetRegister")
        return register

    def __set_motor_enable(self) -> None:
        """Set the enabled state."""
        new_status_word = (
            self.status_word & ~constants.IL_MC_PDS_STA_OE_MSK | constants.IL_MC_PDS_STA_OE
        )
        self.status_word = new_status_word
        self.phasing.enable()

    def __set_motor_disable(self) -> None:
        """Set the disabled state."""
        new_status_word = (
            self.status_word & ~constants.IL_MC_PDS_STA_OE_MSK | constants.IL_MC_PDS_STA_SOD
        )
        self.status_word = new_status_word
        self.phasing.set_phasing_bit()

    def __set_motor_ready_to_switch_on(self) -> None:
        """Set the ready-to-switch-on state."""
        new_status_word = (
            self.status_word & ~constants.IL_MC_PDS_STA_OE_MSK | constants.IL_MC_PDS_STA_RTSO
        )
        self.status_word = new_status_word

    def __emulate_plants(self) -> None:
        """Emulate plants according the operation mode."""
        if self.operation_mode in [OperationMode.VOLTAGE]:
            self._plant_open_loop_rl_d.emulate_plant()
            self._plant_open_loop_rl_q.emulate_plant()
            self._plant_open_loop_vol_to_curr_a.emulate_plant()
            self._plant_open_loop_vol_to_curr_b.emulate_plant()
            self._plant_open_loop_vol_to_curr_c.emulate_plant()
        if self.operation_mode in [OperationMode.CURRENT]:
            self._plant_open_loop_jb.emulate_plant()
            self._plant_closed_loop_rl_d.emulate_plant()
            self._plant_closed_loop_rl_q.emulate_plant()
        if self.operation_mode in [OperationMode.VELOCITY]:
            self._plant_open_loop_position.emulate_plant()
        if self.operation_mode in [
            OperationMode.VELOCITY,
            OperationMode.PROFILE_VELOCITY,
        ]:
            self._plant_closed_loop_jb.emulate_plant()
        if self.operation_mode in [
            OperationMode.POSITION,
            OperationMode.PROFILE_POSITION,
            OperationMode.PROFILE_POSITION_S_CURVE,
        ]:
            self._plant_closed_loop_position.emulate_plant()

    def __clean_plant_signals(self) -> None:
        """Clean all plant signals."""
        self._plant_open_loop_jb.clean_signals()
        self._plant_closed_loop_jb.clean_signals()
        self._plant_open_loop_position.clean_signals()
        self._plant_closed_loop_position.clean_signals()
        self._plant_open_loop_rl_d.clean_signals()
        self._plant_open_loop_rl_q.clean_signals()
        self._plant_closed_loop_rl_d.clean_signals()
        self._plant_closed_loop_rl_q.clean_signals()
        self._plant_open_loop_vol_to_vel.clean_signals()
        self._plant_open_loop_vol_to_curr_a.clean_signals()
        self._plant_open_loop_vol_to_curr_b.clean_signals()
        self._plant_open_loop_vol_to_curr_c.clean_signals()

    @property
    def operation_mode(self) -> int:
        """Operation Mode."""
        return int(self.get_value_by_id(1, "DRV_OP_VALUE"))

    @operation_mode.setter
    def operation_mode(self, operation_mode: int) -> None:
        """Set Operation Mode."""
        self.set_value_by_id(1, "DRV_OP_VALUE", operation_mode)

    @property
    def current_loop_rate(self) -> int:
        """Current loop rate."""
        return int(self.get_value_by_id(1, "CL_CUR_FREQ"))

    @property
    def enabled(self) -> bool:
        """Return true if the motor is enabled.

        Returns:
            True if the motor is enabled.
        """
        return (self.status_word & constants.IL_MC_PDS_STA_OE_MSK) == constants.IL_MC_PDS_STA_OE

    @property
    def status_word(self) -> int:
        return int(self.get_value_by_id(1, "DRV_STATE_STATUS"))

    @status_word.setter
    def status_word(self, value: int) -> None:
        return self.set_value_by_id(1, "DRV_STATE_STATUS", value)



from typing import Optional, List

from ingenialink.dictionary import DictionaryV2, Interface
from ingenialink.canopen.register import CanopenRegister, REG_DTYPE, REG_ACCESS, RegCyclicType

import ingenialogger
import xml.etree.ElementTree as ET

logger = ingenialogger.get_logger(__name__)


class CanopenDictionaryV2(DictionaryV2):
    """Contains all registers and information of a CANopen dictionary.

    Args:
        dictionary_path: Path to the Ingenia dictionary.

    """

    MONITORING_DISTURBANCE_REGISTERS: List[CanopenRegister] = [
        CanopenRegister(
            identifier="MONITORING_DATA",
            idx=0x58B2,
            subidx=0x00,
            cyclic=RegCyclicType.CONFIG,
            dtype=REG_DTYPE.BYTE_ARRAY_512,
            access=REG_ACCESS.RO,
            subnode=0,
        ),
        CanopenRegister(
            identifier="DISTURBANCE_DATA",
            idx=0x58B4,
            subidx=0x00,
            cyclic=RegCyclicType.CONFIG,
            dtype=REG_DTYPE.BYTE_ARRAY_512,
            access=REG_ACCESS.WO,
            subnode=0,
        ),
    ]

    def __init__(self, dictionary_path: str):
        super().__init__(dictionary_path, Interface.CAN)

    def _read_xdf_register(self, register: ET.Element) -> Optional[CanopenRegister]:
        current_read_register = super()._read_xdf_register(register)
        if current_read_register is None:
            return None
        try:
            aux_var = int(register.attrib["address"], 16)
            idx = aux_var >> 8
            subidx = aux_var & 0xFF

            canopen_register = CanopenRegister(
                idx,
                subidx,
                current_read_register.dtype,
                current_read_register.access,
                identifier=current_read_register.identifier,
                units=current_read_register.units,
                cyclic=current_read_register.cyclic,
                phy=current_read_register.phy,
                subnode=current_read_register.subnode,
                storage=current_read_register.storage,
                reg_range=current_read_register.range,
                labels=current_read_register.labels,
                enums=current_read_register.enums,
                cat_id=current_read_register.cat_id,
                scat_id=current_read_register.scat_id,
                internal_use=current_read_register.internal_use,
                address_type=current_read_register.address_type,
            )

            return canopen_register

        except KeyError as ke:
            logger.error(
                f"Register with ID {current_read_register.identifier} has not attribute {ke}"
            )
            return None



import functools
import logging
import struct
import warnings
import xml.etree.ElementTree as ET
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Union

import ingenialogger

from ingenialink.enums.register import REG_DTYPE
from ingenialink.exceptions import ILValueError

if TYPE_CHECKING:
    from ingenialink.servo import Servo


logger = ingenialogger.get_logger(__name__)

POLLING_MAX_TRIES = 5  # Seconds

dtype_value: Dict[REG_DTYPE, Tuple[int, bool]] = {
    REG_DTYPE.U8: (1, False),
    REG_DTYPE.S8: (1, True),
    REG_DTYPE.U16: (2, False),
    REG_DTYPE.S16: (2, True),
    REG_DTYPE.U32: (4, False),
    REG_DTYPE.S32: (4, True),
    REG_DTYPE.U64: (8, False),
    REG_DTYPE.S64: (8, True),
    REG_DTYPE.FLOAT: (4, True),
    REG_DTYPE.BOOL: (1, False),
}

dtype_length_bits: Dict[REG_DTYPE, int] = {
    REG_DTYPE.U8: 8,
    REG_DTYPE.S8: 8,
    REG_DTYPE.U16: 16,
    REG_DTYPE.S16: 16,
    REG_DTYPE.U32: 32,
    REG_DTYPE.S32: 32,
    REG_DTYPE.U64: 64,
    REG_DTYPE.S64: 64,
    REG_DTYPE.FLOAT: 32,
    REG_DTYPE.BOOL: 1,
}

VALID_BIT_REGISTER_VALUES = [0, 1, True, False]


def deprecated(
    custom_msg: Optional[str] = None, new_func_name: Optional[str] = None
) -> Callable[..., Any]:
    """This is a decorator which can be used to mark functions as deprecated.
    It will result in a warning being emitted when the function is used. We use
    this decorator instead of any deprecation library because all libraries raise
    a DeprecationWarning but since by default this warning is hidden, we use this
    decorator to manually activate DeprecationWarning and turning it off after
    the warn has been done."""

    def wrap(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapped_method(*args: Any, **kwargs: Any) -> Any:
            warnings.simplefilter("always", DeprecationWarning)  # Turn off filter
            msg = 'Call to deprecated function "{}".'.format(func.__name__)
            if new_func_name:
                msg += ' Please, use "{}" function instead.'.format(new_func_name)
            if custom_msg:
                msg = custom_msg
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
            warnings.simplefilter("ignore", DeprecationWarning)  # Reset filter
            return func(*args, **kwargs)

        return wrapped_method

    return wrap


class DisableLogger:
    """Context manager to disable all logs."""

    def __enter__(self) -> None:
        logging.disable(logging.CRITICAL)

    def __exit__(self, *args: Any) -> None:
        logging.disable(logging.NOTSET)


def to_ms(s: Union[int, float]) -> int:
    """Convert from seconds to milliseconds.

    Args:
        s: Value in seconds.

    Returns:
        Value in milliseconds.
    """
    return int(s * 1e3)


def remove_xml_subelement(element: ET.Element, subelement: ET.Element) -> None:
    """Removes a subelement from the given element the element contains the subelement

    Args:
        element: Element to be extracted from.
        subelement: Element to be extracted.
    """
    if subelement is not None and subelement in element:
        element.remove(subelement)


def pop_element(dictionary: Dict[str, Any], element: str) -> None:
    """Pops an element from a dictionary only if it is contained in it

    Args:
        dictionary: Dictionary containing all the elements
        element: Element to be popped from the dictionary.
    """
    if element in dictionary:
        dictionary.pop(element)


def cleanup_register(register: ET.Element) -> None:
    """Cleans a ElementTree register to remove all
    unnecessary fields for a configuration file

    Args:
        register: Register to be cleaned.
    """
    labels = register.find("./Labels")
    range = register.find("./Range")
    enums = register.find("./Enumerations")

    if labels:
        remove_xml_subelement(register, labels)
    if enums:
        remove_xml_subelement(register, enums)
    if range:
        remove_xml_subelement(register, range)

    pop_element(register.attrib, "desc")
    pop_element(register.attrib, "cat_id")
    pop_element(register.attrib, "cyclic")
    pop_element(register.attrib, "units")
    pop_element(register.attrib, "address_type")

    register.text = ""


def get_drive_identification(
    servo: "Servo", subnode: Optional[int] = None
) -> Tuple[Optional[int], Optional[int]]:
    """Gets the identification information of a given subnode.

    Args:
        servo: Instance of the servo Class.
        subnode: subnode to be targeted.

    Returns:
        int, Product code and revision number of the targeted subnode.
    """
    prod_code = None
    re_number = None
    try:
        if subnode is None or subnode == 0:
            prod_code = int(servo.read("DRV_ID_PRODUCT_CODE_COCO", 0))
            re_number = int(servo.read("DRV_ID_REVISION_NUMBER_COCO", 0))
        else:
            prod_code = int(servo.read("DRV_ID_PRODUCT_CODE", subnode=subnode))
            re_number = int(servo.read("DRV_ID_REVISION_NUMBER", subnode))
    except Exception as e:
        pass

    return prod_code, re_number


def convert_ip_to_int(ip: str) -> int:
    """Converts a string type IP to its integer value.

    Args:
        ip: IP to be converted.

    """
    split_ip = ip.split(".")
    drive_ip1 = int(split_ip[0]) << 24
    drive_ip2 = int(split_ip[1]) << 16
    drive_ip3 = int(split_ip[2]) << 8
    drive_ip4 = int(split_ip[3])
    return drive_ip1 + drive_ip2 + drive_ip3 + drive_ip4


def convert_int_to_ip(int_ip: int) -> str:
    """Converts an integer type IP to its string form.

    Args:
        int_ip: IP to be converted.

    """
    drive_ip1 = (int_ip >> 24) & 0x000000FF
    drive_ip2 = (int_ip >> 16) & 0x000000FF
    drive_ip3 = (int_ip >> 8) & 0x000000FF
    drive_ip4 = int_ip & 0x000000FF
    return "{}.{}.{}.{}".format(drive_ip1, drive_ip2, drive_ip3, drive_ip4)


class INT_SIZES(Enum):
    """Integer sizes."""

    S8_MIN = -128
    S16_MIN = -32767 - 1
    S32_MIN = -2147483647 - 1
    S64_MIN = 9223372036854775807 - 1

    S8_MAX = 127
    S16_MAX = 32767
    S32_MAX = 2147483647
    S64_MAX = 9223372036854775807

    U8_MAX = 255
    U16_MAX = 65535
    U32_MAX = 4294967295
    U64_MAX = 18446744073709551615


def convert_bytes_to_dtype(data: bytes, dtype: REG_DTYPE) -> Union[float, int, str]:
    """Convert data in bytes to corresponding dtype.

    Args:
        data: data to convert
        dtype: output dtype

    Raises:
        ILValueError: If data can't be decoded in utf-8
    """
    signed = False
    if dtype in dtype_value:
        bytes_length, signed = dtype_value[dtype]
        data = data[:bytes_length]

    if dtype == REG_DTYPE.FLOAT:
        [value] = struct.unpack("f", data)
        if not isinstance(value, float):
            raise ILValueError(f"Data could not be converted to float. Obtained: {value}")
    elif dtype == REG_DTYPE.STR:
        try:
            value = data.split(b"\x00")[0].decode("utf-8")
        except UnicodeDecodeError as e:
            raise ILValueError(f"Can't decode {e.object!r} to utf-8 string") from e
    else:
        value = int.from_bytes(data, "little", signed=signed)
    if dtype == REG_DTYPE.BOOL:
        value = bool(value)
    if not isinstance(value, (int, float, str)):
        raise ILValueError(f"Bad data type: {type(value)}")
    return value


def convert_dtype_to_bytes(data: Union[int, float, str, bytes], dtype: REG_DTYPE) -> bytes:
    """Convert data in dtype to bytes.
    Args:
        data: Data to convert.
        dtype: Data type.
    """
    if (
        dtype == REG_DTYPE.BOOL
        and data not in VALID_BIT_REGISTER_VALUES
        and not isinstance(data, bytes)
    ):
        raise ValueError(f"Invalid value. Expected values: {VALID_BIT_REGISTER_VALUES}, got {data}")
    if dtype == REG_DTYPE.BYTE_ARRAY_512:
        if not isinstance(data, bytes):
            raise ValueError(f"Expected data of type bytes, but got {type(data)}")
        return data
    if dtype == REG_DTYPE.FLOAT:
        if not isinstance(data, (float, int)):
            raise ValueError(f"Expected data of type float, but got {type(data)}")
        return struct.pack("f", float(data))
    if dtype == REG_DTYPE.STR:
        if not isinstance(data, str):
            raise ValueError(f"Expected data of type string, but  got {type(data)}")
        return data.encode("utf_8")
    if not isinstance(data, int):
        raise ValueError(f"Expected data of type int, but {type(data)}")
    bytes_length, signed = dtype_value[dtype]
    data_bytes = data.to_bytes(bytes_length, byteorder="little", signed=signed)
    return data_bytes



import pytest

from ingenialink.exceptions import ILValueError
from ingenialink.enums.register import REG_DTYPE
from ingenialink.utils._utils import convert_bytes_to_dtype


@pytest.mark.parametrize(
    "v_input, v_output, dtype",
    [
        (b"\x00", 0, REG_DTYPE.S32),
        (b"\x00\x00\x0a\x42", 34.5, REG_DTYPE.FLOAT),
        (b"\xFF", 255, REG_DTYPE.U16),
        (b"\x74\x68\x61\x74\x27\x73\x20\x61\x20\x74\x65\x73\x74", "that's a test", REG_DTYPE.STR),
        (b"\x74\x68\x61\x74\x27\x73\x20\x67\x6f\x6f\x64\x00\xca\xca", "that's good", REG_DTYPE.STR),
    ],
)
def test_convert_bytes_to_dtype(v_input, v_output, dtype):
    assert convert_bytes_to_dtype(v_input, dtype) == v_output


def test_convert_bytes_to_dtype_wrong_string():
    wrong_data = b"\xff\xff\xff\xff\xff\x00"
    with pytest.raises(ILValueError):
        convert_bytes_to_dtype(wrong_data, REG_DTYPE.STR)



from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, List, Union

import ingenialogger

from ingenialink.servo import Servo

logger = ingenialogger.get_logger(__name__)


class NET_PROT(Enum):
    """Network Protocol."""

    EUSB = 0
    MCB = 1
    ETH = 2
    ECAT = 3
    CAN = 5


class NET_STATE(Enum):
    """Network State."""

    CONNECTED = 0
    DISCONNECTED = 1
    FAULTY = 2


class NET_DEV_EVT(Enum):
    """Device Event."""

    ADDED = 0
    REMOVED = 1


class EEPROM_FILE_FORMAT(Enum):
    """EEPROM file format"""

    BINARY = 0
    INTEL = 1


class NET_TRANS_PROT(Enum):
    """Transmission protocol."""

    TCP = 1
    UDP = 2


@dataclass
class SlaveInfo:
    product_code: int
    revision_number: int


class Network(ABC):
    """Declaration of a general Network object."""

    def __init__(self) -> None:
        self.servos: List[Any] = []
        """List of the connected servos in the network."""

    @abstractmethod
    def scan_slaves(self) -> List[int]:
        raise NotImplementedError

    @abstractmethod
    def scan_slaves_info(self) -> OrderedDict[int, SlaveInfo]:
        raise NotImplementedError

    @abstractmethod
    def connect_to_slave(self, *args: Any, **kwargs: Any) -> Servo:
        raise NotImplementedError

    @abstractmethod
    def disconnect_from_slave(self, servo: Servo) -> None:
        raise NotImplementedError

    @abstractmethod
    def load_firmware(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    @abstractmethod
    def subscribe_to_status(
        self, target: Union[int, str], callback: Callable[[NET_DEV_EVT], Any]
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def unsubscribe_from_status(
        self, target: Union[int, str], callback: Callable[[NET_DEV_EVT], Any]
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def start_status_listener(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    @abstractmethod
    def stop_status_listener(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    @property
    def protocol(self) -> NET_PROT:
        raise NotImplementedError



import ipaddress
import socket
import time
from collections import OrderedDict
from enum import Enum
from typing import Dict, List, Optional

import ingenialogger

from ingenialink import constants
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.ethernet.servo import EthernetServo
from ingenialink.exceptions import ILError, ILIOError, ILTimeoutError
from ingenialink.network import SlaveInfo

logger = ingenialogger.get_logger(__name__)


class EoECommand(Enum):
    INIT = 0
    DEINIT = 1
    SCAN = 2
    CONFIG = 3
    ERASE_CONFIG = 4
    EOE_START = 5
    EOE_STOP = 6
    GET_STATUS = 7


class EoENetwork(EthernetNetwork):
    """Network for EoE (Ethernet over EtherCAT) communication.

    Args:
        ifname: Network interface name.
        connection_timeout: Time in seconds of the connection timeout to the EoE service.

    """

    EOE_MSG_CMD_SIZE = 2
    EOE_MSG_NODE_SIZE = 2
    EOE_MSG_IP_SIZE = 4
    EOE_MSG_DATA_SIZE = 53
    EOE_MSG_FRAME_SIZE = EOE_MSG_CMD_SIZE + EOE_MSG_DATA_SIZE
    NULL_TERMINATOR = b"\x00"

    STATUS_EOE_BIT = 0b10
    STATUS_INIT_BIT = 0b1
    WAIT_EOE_TIMEOUT = 1

    ECAT_SERVICE_NETWORK = ipaddress.ip_network("192.168.3.0/24")

    def __init__(
        self, ifname: str, connection_timeout: float = constants.DEFAULT_ETH_CONNECTION_TIMEOUT
    ) -> None:
        super().__init__()
        self.ifname = ifname
        self._eoe_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._eoe_socket.settimeout(connection_timeout)
        self._connect_to_eoe_service()
        status = self._get_status_eoe_service()
        if status & self.STATUS_EOE_BIT:
            self._stop_eoe_service()
            self._erase_config_eoe_service()
        if status & self.STATUS_INIT_BIT:
            self._deinitialize_eoe_service()
        self._eoe_service_init = False
        self._eoe_service_started = False
        self._configured_slaves: Dict[str, int] = {}

    def connect_to_slave(  # type: ignore [override]
        self,
        slave_id: int,
        ip_address: str,
        dictionary: str,
        port: int = 1061,
        connection_timeout: float = constants.DEFAULT_ETH_CONNECTION_TIMEOUT,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> EthernetServo:
        """Connects to a slave through the given network settings.

        Args:
            slave_id: EtherCAT slave ID.
            ip_address: IP address to be assigned to the slave.
            dictionary: Path to the target dictionary file.
            port: Port to connect to the slave.
            connection_timeout: Time in seconds of the connection timeout.
            servo_status_listener: Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener: Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            ValueError: ip_address must be a subnetwork of 192.168.3.0/24
            ILError: If the EoE service is not running.
            ILError: If the EoE service cannot be started on the network interface.

        Returns:
            EthernetServo: Instance of the servo connected.

        """
        if ipaddress.ip_address(ip_address) not in self.ECAT_SERVICE_NETWORK:
            raise ValueError("ip_address must be a subnetwork of 192.168.3.0/24")
        if not self._eoe_service_init:
            self._initialize_eoe_service()
        if self._eoe_service_started:
            self._stop_eoe_service()
            self._erase_config_eoe_service()
        self.__reconfigure_drives()
        try:
            self._configure_slave(slave_id, ip_address)
        finally:
            self._start_eoe_service()
        self.__wait_eoe_starts()
        self._configured_slaves[ip_address] = slave_id
        return super().connect_to_slave(
            ip_address,
            dictionary,
            port,
            connection_timeout,
            servo_status_listener,
            net_status_listener,
            True,
        )

    def __wait_eoe_starts(self) -> None:
        """Wait until the EoE service starts the EoE or the timeout was reached"""
        status = self._get_status_eoe_service()
        time_start = time.time()
        while not status & self.STATUS_EOE_BIT and self.WAIT_EOE_TIMEOUT > time.time() - time_start:
            time.sleep(0.1)
            status = self._get_status_eoe_service()
        if not status & self.STATUS_EOE_BIT:
            logger.warning("Service did not starts the EoE")

    def __reconfigure_drives(self) -> None:
        """Reconfigure all the slaves saved in the network"""
        for ip_addr, slave_id in self._configured_slaves.items():
            try:
                self._configure_slave(slave_id, ip_addr)
            except ILError as e:
                logger.error(e)

    def disconnect_from_slave(self, servo: EthernetServo) -> None:  # type: ignore [override]
        del self._configured_slaves[servo.ip_address]
        super().disconnect_from_slave(servo)
        if len(self.servos) == 0:
            self._stop_eoe_service()
            self._erase_config_eoe_service()
            try:
                self._deinitialize_eoe_service()
            except ILError as e:
                logger.error(e)

    def __del__(self) -> None:
        self._eoe_socket.shutdown(socket.SHUT_RDWR)
        self._eoe_socket.close()

    def scan_slaves(self) -> List[int]:
        """Scan slaves connected to the network adapter.

        Returns:
            List containing the ids of the connected slaves.

        Raises:
            ILError: If the EoE service fails to perform a scan.

        """
        deinit_later = False
        if not self._eoe_service_init:
            deinit_later = True
            self._initialize_eoe_service()
        result = self._scan_eoe_service()
        if deinit_later:
            self._deinitialize_eoe_service()
        return result

    def _scan_eoe_service(self) -> List[int]:
        """Make the scan request to the EoE service

        Returns:
            List containing the ids of the connected slaves.

        Raises:
            ILError: If the EoE service fails to perform a scan.

        """
        msg = self._build_eoe_command_msg(EoECommand.SCAN.value)
        try:
            r = self._send_command(msg)
        except (ILIOError, ILTimeoutError) as e:
            raise ILError(
                "Failed to perform a network scan. Please verify the EoE service is running."
            ) from e
        return list(range(1, r + 1))

    def scan_slaves_info(self) -> OrderedDict[int, SlaveInfo]:
        raise NotImplementedError

    @staticmethod
    def _build_eoe_command_msg(cmd: int, data: Optional[bytes] = None) -> bytes:
        """
        Build a message with the following format.

        +----------+----------+
        |   cmd    |   data   |
        +==========+==========+
        |  2 Byte  | 53 Bytes |
        +----------+----------+

        Args:
            cmd: Indicates which operation to perform.
            data: Contains the necessary data to perform the desired command.

        Returns:
            The message to send.

        """
        if data is None:
            data = bytes()
        cmd_field = cmd.to_bytes(EoENetwork.EOE_MSG_CMD_SIZE, "little")
        data_field = data + EoENetwork.NULL_TERMINATOR * (EoENetwork.EOE_MSG_DATA_SIZE - len(data))
        return cmd_field + data_field

    def _send_command(self, msg: bytes) -> int:
        """
        Send command to EoE service.

        Args:
            msg: Message to send.

        Returns:
            Response from the EoE service.

        Raises:
            ILTimeoutError: Timeout while receiving a response from
            the EoE service.
            ILIOError: Error while sending/receiving message.

        """
        try:
            self._eoe_socket.send(msg)
        except socket.error as e:
            raise ILIOError("Error sending message.") from e
        try:
            response = self._eoe_socket.recv(1024)
        except socket.timeout as e:
            raise ILTimeoutError("Timeout while receiving response.") from e
        except socket.error as e:
            raise ILIOError("Error receiving response.") from e
        return int.from_bytes(response, byteorder="little", signed=True)

    def _connect_to_eoe_service(self) -> None:
        """Connect to the EoE service."""
        self._eoe_socket.connect(("127.0.0.1", 8888))

    def _initialize_eoe_service(self) -> None:
        """Initialize the virtual network interface and the packet forwarder.

        Raises:
            ILError: If the EoE service is not running.
            ILError: If the EoE service cannot be started on the network interface.

        """
        self._eoe_service_init = True
        data = self.ifname
        msg = self._build_eoe_command_msg(EoECommand.INIT.value, data=data.encode("utf-8"))
        try:
            r = self._send_command(msg)
        except (ILIOError, ILTimeoutError) as e:
            raise ILError(
                "Failed to initialize the EoE service. Please verify it's running."
            ) from e
        if r < 0:
            raise ILError(f"Failed to initialize the EoE service using interface {self.ifname}.")

    def _deinitialize_eoe_service(self) -> None:
        """Deinitialize the virtual network interface and the packet forwarder.

        Raises:
            ILError: If the EoE service is not running.
            ILError: If the EoE service cannot be stopped on the network interface.

        """
        data = self.ifname
        msg = self._build_eoe_command_msg(EoECommand.DEINIT.value, data=data.encode("utf-8"))
        try:
            self._send_command(msg)
            self._eoe_service_init = False
        except (ILIOError, ILTimeoutError) as e:
            raise ILError("Failed to deinitialize the EoE service.") from e

    def _configure_slave(
        self, slave_id: int, ip_address: str, net_mask: str = "255.255.255.0"
    ) -> None:
        """
        Configure an EtherCAT slave with a given IP.

        Args:
            slave_id: EtherCAT slave ID.
            ip_address: IP address to be set to the slave.

        Raises:
            ILError: If the EoE service fails to configure a slave.

        """
        slave_bytes = slave_id.to_bytes(self.EOE_MSG_NODE_SIZE, "little")
        ip_int = int(ipaddress.IPv4Address(ip_address))
        ip_bytes = ip_int.to_bytes(self.EOE_MSG_IP_SIZE, "little")
        net_mask_int = int(ipaddress.IPv4Address(net_mask))
        net_mask_bytes = net_mask_int.to_bytes(self.EOE_MSG_IP_SIZE, "little")
        data = slave_bytes + ip_bytes + net_mask_bytes
        msg = self._build_eoe_command_msg(EoECommand.CONFIG.value, data)
        try:
            self._send_command(msg)
        except (ILIOError, ILTimeoutError) as e:
            raise ILError(f"Failed to configure slave {slave_id} with IP {ip_address}.") from e

    def _start_eoe_service(self) -> None:
        """Starts the EoE service

        Raises:
           ILError: If the EoE service fails to start.

        """
        self._eoe_service_started = True
        msg = self._build_eoe_command_msg(EoECommand.EOE_START.value)
        try:
            self._send_command(msg)
        except (ILIOError, ILTimeoutError) as e:
            raise ILError("Failed to start the EoE service.") from e

    def _stop_eoe_service(self) -> None:
        """Stops the EoE service

        Raises:
           ILError: If the EoE service fails to stop.

        """
        self._eoe_service_started = False
        msg = self._build_eoe_command_msg(EoECommand.EOE_STOP.value)
        try:
            self._send_command(msg)
        except (ILIOError, ILTimeoutError) as e:
            raise ILError("Failed to stop the EoE service.") from e

    def _erase_config_eoe_service(self) -> None:
        """Stops the EoE service

        Raises:
           ILError: If the EoE service fails to stop.

        """
        msg = self._build_eoe_command_msg(EoECommand.ERASE_CONFIG.value)
        try:
            self._send_command(msg)
        except (ILIOError, ILTimeoutError) as e:
            raise ILError("Failed to stop the EoE service.") from e

    def _get_status_eoe_service(self) -> int:
        """Get the EoE service status.

        +-----------+------+------+
        |ECAT status| init | eoe  |
        +===========+======+======+
        |   1Byte   | 1bit | 1bit |
        +-----------+------+------+

        Returns:
            Status response

        Raises:
           ILError: If get status request fails.

        """
        msg = self._build_eoe_command_msg(EoECommand.GET_STATUS.value)
        try:
            r = self._send_command(msg)
        except (ILIOError, ILTimeoutError) as e:
            raise ILError("Failed to get service status.") from e
        return r

    def load_firmware_moco(self) -> None:  # type: ignore [override]
        raise NotImplementedError

    def load_firmware(self) -> None:  # type: ignore [override]
        raise NotImplementedError



from typing import Optional, List
import xml.etree.ElementTree as ET

import ingenialogger

from ingenialink.dictionary import DictionaryV2, Interface
from ingenialink.ethercat.register import (
    EthercatRegister,
    REG_DTYPE,
    REG_ACCESS,
    REG_ADDRESS_TYPE,
    RegCyclicType,
)
from ingenialink.constants import (
    CANOPEN_ADDRESS_OFFSET,
    CANOPEN_SUBNODE_0_ADDRESS_OFFSET,
    MAP_ADDRESS_OFFSET,
)

logger = ingenialogger.get_logger(__name__)


class EthercatDictionaryV2(DictionaryV2):
    """Contains all registers and information of a EtherCAT dictionary.

    Args:
        dictionary_path: Path to the Ingenia dictionary.

    """

    MONITORING_DISTURBANCE_REGISTERS: List[EthercatRegister] = [
        EthercatRegister(
            identifier="MONITORING_DATA",
            units="",
            subnode=0,
            idx=0x58B2,
            subidx=0x01,
            cyclic=RegCyclicType.CONFIG,
            dtype=REG_DTYPE.BYTE_ARRAY_512,
            access=REG_ACCESS.RO,
        ),
        EthercatRegister(
            identifier="DISTURBANCE_DATA",
            units="",
            subnode=0,
            idx=0x58B4,
            subidx=0x01,
            cyclic=RegCyclicType.CONFIG,
            dtype=REG_DTYPE.BYTE_ARRAY_512,
            access=REG_ACCESS.WO,
        ),
    ]

    PDO_REGISTERS: List[EthercatRegister] = [
        EthercatRegister(
            identifier="RPDO_ASSIGN_REGISTER_SUB_IDX_0",
            units="",
            subnode=0,
            idx=0x1C12,
            subidx=0x00,
            dtype=REG_DTYPE.S32,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
        EthercatRegister(
            identifier="RPDO_ASSIGN_REGISTER_SUB_IDX_1",
            units="",
            subnode=0,
            idx=0x1C12,
            subidx=0x01,
            dtype=REG_DTYPE.S32,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
        EthercatRegister(
            identifier="RPDO_MAP_REGISTER_SUB_IDX_0",
            units="",
            subnode=0,
            idx=0x1600,
            subidx=0x00,
            dtype=REG_DTYPE.S32,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
        EthercatRegister(
            identifier="RPDO_MAP_REGISTER_SUB_IDX_1",
            units="",
            subnode=0,
            idx=0x1600,
            subidx=0x01,
            dtype=REG_DTYPE.STR,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
        EthercatRegister(
            identifier="TPDO_ASSIGN_REGISTER_SUB_IDX_0",
            units="",
            subnode=0,
            idx=0x1C13,
            subidx=0x00,
            dtype=REG_DTYPE.S32,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
        EthercatRegister(
            identifier="TPDO_ASSIGN_REGISTER_SUB_IDX_1",
            units="",
            subnode=0,
            idx=0x1C13,
            subidx=0x01,
            dtype=REG_DTYPE.S32,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
        EthercatRegister(
            identifier="TPDO_MAP_REGISTER_SUB_IDX_0",
            units="",
            subnode=0,
            idx=0x1A00,
            subidx=0x00,
            dtype=REG_DTYPE.S32,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
        EthercatRegister(
            identifier="TPDO_MAP_REGISTER_SUB_IDX_1",
            units="",
            subnode=0,
            idx=0x1A00,
            subidx=0x01,
            dtype=REG_DTYPE.STR,
            access=REG_ACCESS.RW,
            address_type=REG_ADDRESS_TYPE.NVM_NONE,
        ),
    ]

    def __init__(self, dictionary_path: str):
        super().__init__(dictionary_path, Interface.ECAT)

    @staticmethod
    def __get_cia_offset(subnode: int) -> int:
        """Get the CiA offset for the register based on the subnode.

        Args:
            subnode: register subnode.

        Returs:
            The CiA offset for the register.

        """
        return (
            CANOPEN_SUBNODE_0_ADDRESS_OFFSET
            if subnode == 0
            else CANOPEN_ADDRESS_OFFSET + MAP_ADDRESS_OFFSET * (subnode - 1)
        )

    def _read_xdf_register(self, register: ET.Element) -> Optional[EthercatRegister]:
        current_read_register = super()._read_xdf_register(register)
        if current_read_register is None:
            return None
        try:
            idx = int(register.attrib["address"], 16) + self.__get_cia_offset(
                current_read_register.subnode
            )
            subidx = 0x00

            ethercat_register = EthercatRegister(
                idx,
                subidx,
                current_read_register.dtype,
                current_read_register.access,
                identifier=current_read_register.identifier,
                units=current_read_register.units,
                cyclic=current_read_register.cyclic,
                phy=current_read_register.phy,
                subnode=current_read_register.subnode,
                storage=current_read_register.storage,
                reg_range=current_read_register.range,
                labels=current_read_register.labels,
                enums=current_read_register.enums,
                cat_id=current_read_register.cat_id,
                scat_id=current_read_register.scat_id,
                internal_use=current_read_register.internal_use,
                address_type=current_read_register.address_type,
            )

            return ethercat_register

        except KeyError as ke:
            logger.error(
                f"Register with ID {current_read_register.identifier} has not attribute {ke}"
            )
            return None

    def _append_missing_registers(
        self,
    ) -> None:
        """Append missing registers to the dictionary.

        Mainly registers needed for Monitoring/Disturbance and PDOs.

        """
        super()._append_missing_registers()
        for register in self.PDO_REGISTERS:
            if register.identifier is not None:
                self._registers[register.subnode][register.identifier] = register



from typing import List, Optional, Union, Dict

import bitarray

from ingenialink.canopen.register import CanopenRegister
from ingenialink.enums.register import REG_DTYPE, REG_ACCESS, RegCyclicType
from ingenialink.ethercat.register import EthercatRegister
from ingenialink.exceptions import ILError
from ingenialink.servo import Servo
from ingenialink.utils._utils import (
    convert_bytes_to_dtype,
    convert_dtype_to_bytes,
    dtype_length_bits,
)

BIT_ENDIAN = "little"
bitarray._set_default_endian(BIT_ENDIAN)

PADDING_REGISTER_IDENTIFIER = "PADDING"


class PDOMapItem:
    """Abstract class to represent a register in the PDO mapping.

    Attributes:
        register: mapped register object. If None the item will padding.
        size_bits: custom register size in bits.

    Raises:
        ValueError: If the register and size_bits are not provided.
        ValueError: If the size_bits value is invalid. Only when the register
        is set to None.

    """

    ACCEPTED_CYCLIC: RegCyclicType
    """Accepted cyclic: CYCLIC_TX, CYCLIC_RX or CYCLIC_TXRX."""

    def __init__(
        self,
        register: Union[None, EthercatRegister, CanopenRegister] = None,
        size_bits: Optional[int] = None,
    ) -> None:
        if register is None:
            if size_bits is None:
                raise ValueError("The size bits must be set when creating padding items.")
            register = EthercatRegister(
                identifier=PADDING_REGISTER_IDENTIFIER,
                units="",
                subnode=0,
                idx=0x0000,
                subidx=0x00,
                cyclic=self.ACCEPTED_CYCLIC,
                dtype=REG_DTYPE.STR,
                access=REG_ACCESS.RW,
            )
        self.register = register
        self.size_bits = size_bits or dtype_length_bits[register.dtype]
        self._raw_data_bits: Optional[bitarray.bitarray] = None
        self._check_if_mappable()

    def _check_if_mappable(self) -> None:
        """Check if the passed register is mappable. I.e., if the cyclic information is correct.

        Raises:
            ILError: Tf the register is not mappable.
        """
        if self.register.cyclic not in [self.ACCEPTED_CYCLIC, RegCyclicType.TXRX]:
            raise ILError(
                f"Incorrect cyclic. It should be {self.ACCEPTED_CYCLIC} or {RegCyclicType.TXRX},"
                f" obtained: {self.register.cyclic}"
            )

    @property
    def raw_data_bits(self) -> bitarray.bitarray:
        """Raw data in bits.

        Returns:
            Raw data in bits

        Raises:
            ILError: If the raw data is empty.

        """
        if self._raw_data_bits is None:
            raise ILError("Raw data is empty.")
        return self._raw_data_bits

    @raw_data_bits.setter
    def raw_data_bits(self, data: bitarray.bitarray) -> None:
        if len(data) != self.size_bits:
            raise ILError(f"Wrong size. Expected {self.size_bits}, obtained {len(data)}")
        if data.endian() != BIT_ENDIAN:
            raise ILError("Bitarray should be little endian.")
        self._raw_data_bits = data

    @property
    def raw_data_bytes(self) -> bytes:
        """Raw data in bytes.

        Returns:
            Raw data in bytes

        Raises:
            ILError: If the raw data is empty.

        """
        if self._raw_data_bits is None:
            raise ILError("Raw data is empty.")
        return self._raw_data_bits.tobytes()

    @raw_data_bytes.setter
    def raw_data_bytes(self, data: bytes) -> None:
        data_bits = bitarray.bitarray(endian=BIT_ENDIAN)
        data_bits.frombytes(data)
        if self.register.identifier == PADDING_REGISTER_IDENTIFIER:
            data_bits = data_bits[: self.size_bits]
        self.raw_data_bits = data_bits

    @property
    def value(self) -> Union[int, float, bool]:
        """Register value. Converts the raw data bytes into the register value.

        Raises:
            ILError: If the raw data is empty.
            ILError: If the register type is not int or float.

        Returns:
            Register value.
        """
        value: Union[bool, int, float, str]
        if self.register.identifier == PADDING_REGISTER_IDENTIFIER:
            raise NotImplementedError(
                "The register value must be read by the raw_data_bytes attribute."
            )
        if self.register.dtype == REG_DTYPE.BOOL:
            value = self.raw_data_bits.any()
        else:
            value = convert_bytes_to_dtype(self.raw_data_bytes, self.register.dtype)
        if not isinstance(value, (int, float, bool)):
            raise ILError("Wrong register value type")
        return value

    @property
    def register_mapping(self) -> bytes:
        """Arrange register information into PDO mapping format.

        Returns:
            PDO register mapping format.

        """
        index = self.register.idx
        mapped_register = (index << 16) | self.size_bits
        mapped_register_bytes: bytes = mapped_register.to_bytes(4, "little")
        return mapped_register_bytes


class RPDOMapItem(PDOMapItem):
    """Class to represent RPDO mapping items."""

    ACCEPTED_CYCLIC = RegCyclicType.RX

    def __init__(
        self,
        register: Union[None, EthercatRegister, CanopenRegister] = None,
        size_bits: Optional[int] = None,
    ) -> None:
        super().__init__(register, size_bits)

    @property
    def value(self) -> Union[int, float]:
        return super().value

    @value.setter
    def value(self, value: Union[int, float, bool]) -> None:
        if self.register.identifier == PADDING_REGISTER_IDENTIFIER:
            raise NotImplementedError(
                "The register value must be set by the raw_data_bytes attribute."
            )
        if isinstance(value, bool):
            raw_data_bits = bitarray.bitarray(endian=BIT_ENDIAN)
            raw_data_bits.append(value)
            self.raw_data_bits = raw_data_bits
        else:
            raw_data_bytes = convert_dtype_to_bytes(value, self.register.dtype)
            self.raw_data_bytes = raw_data_bytes


class TPDOMapItem(PDOMapItem):
    """Class to represent TPDO mapping items."""

    ACCEPTED_CYCLIC = RegCyclicType.TX


class PDOMap:
    """Abstract class that contains PDO mapping information."""

    _PDO_MAP_ITEM_CLASS = PDOMapItem

    def __init__(self) -> None:
        self.__items: List[PDOMapItem] = []
        self.__map_register_address: Optional[int] = None

    def create_item(
        self, register: Union[EthercatRegister, CanopenRegister], size_bits: Optional[int] = None
    ) -> PDOMapItem:
        """Create a new PDOMapItem.

        Args:
            register: Register object.
            size_bits: Register size in bits.

        Returns:
            PDO Map item.
        """
        item = self._PDO_MAP_ITEM_CLASS(register, size_bits)
        return item

    def add_item(self, item: PDOMapItem) -> None:
        """Append a new item.

        Args:
            item: Item to be added.
        """
        self.__items.append(item)

    def add_registers(
        self,
        registers: Union[
            Union[EthercatRegister, CanopenRegister], List[Union[EthercatRegister, CanopenRegister]]
        ],
    ) -> None:
        """Add a register or a list of registers in bulk.

        It creates a new item for each register and adds it the items attribute.

        Args:
            registers: Register object or list of Registers.
        """
        if not isinstance(registers, list):
            registers = [registers]
        for register in registers:
            item = self.create_item(register)
            self.add_item(item)

    @property
    def items(self) -> List[PDOMapItem]:
        """List of items.

        Returns:
            List of items.
        """
        return self.__items

    @property
    def map_register_index_bytes(self) -> bytes:
        """Index of the mapping register in bytes.

        Returns:
            Index of the mapping register in bytes.

        Raises:
            ValueError: If map_register_index is None
        """
        if self.map_register_index is None:
            raise ValueError("map_register_index is None")
        else:
            return self.map_register_index.to_bytes(4, "little")

    @property
    def map_register_index(self) -> Optional[int]:
        """Index of the mapping register. None if it is not mapped in the slave.

        Returns:
            Index of the mapping register.
        """
        return self.__map_register_address

    @map_register_index.setter
    def map_register_index(self, address: int) -> None:
        self.__map_register_address = address

    @property
    def data_length_bits(self) -> int:
        """Length of the map in bits.

        Returns:
            Length of the map in bits.
        """
        return sum(item.size_bits for item in self.items)

    @property
    def data_length_bytes(self) -> int:
        """Length of the map in bytes.

        Returns:
            Length of the map in bytes.
        """
        return bitarray.bits2bytes(self.data_length_bits)

    @property
    def items_mapping(self) -> bytearray:
        """Returns all register item mappings concatenated.

        Returns:
            int: _description_
        """
        map_bytes = bytearray()
        for pdo_map_item in self.items:
            map_bytes += pdo_map_item.register_mapping
        return map_bytes


class RPDOMap(PDOMap):
    """Class to store RPDO mapping information."""

    _PDO_MAP_ITEM_CLASS = RPDOMapItem

    def get_item_bits(self) -> bitarray.bitarray:
        """Return the concatenated items raw data to be sent to the slave (in bits).

        Raises:
            ILError: Raw data is empty.
            ILError: If the length of the bit array is incorrect.

        Returns:
            Concatenated items raw data in bits.
        """
        data_bits = bitarray.bitarray(endian=BIT_ENDIAN)
        for item in self.items:
            try:
                data_bits += item.raw_data_bits
            except ILError:
                raise ILError(f"PDO item {item.register.identifier} does not have data stored.")

        if len(data_bits) != self.data_length_bits:
            raise ILError(
                "The length in bits of the data array is incorrect. Expected"
                f" {self.data_length_bits}, obtained {len(data_bits)}"
            )
        return data_bits

    def get_item_bytes(self) -> bytes:
        """Return the concatenated items raw data to be sent to the slave (in bytes).

        Raises:
            ILError: Raw data is empty.
            ILError: If the length of th byte array is incorrect.

        Returns:
            Concatenated items raw data in bytes.
        """
        item_bits = self.get_item_bits()
        return item_bits.tobytes()


class TPDOMap(PDOMap):
    """Class to store TPDO mapping information."""

    _PDO_MAP_ITEM_CLASS = TPDOMapItem

    def set_item_bytes(self, data_bytes: bytes) -> None:
        """Set the items raw data from a byte array received from the slave.

        Args:
            data_bytes: Byte array received from the slave.

        Raises:
            ILError: If the length of the received data does not coincide.
        """
        if len(data_bytes) != self.data_length_bytes:
            raise ILError(
                f"The length of the data array is incorrect. Expected {self.data_length_bytes},"
                f" obtained {len(data_bytes)}"
            )
        data_bits = bitarray.bitarray(endian=BIT_ENDIAN)
        data_bits.frombytes(data_bytes)

        offset = 0
        for item in self.items:
            item.raw_data_bits = data_bits[offset : item.size_bits + offset]
            offset += item.size_bits


class PDOServo(Servo):
    """Abstract class to implement PDOs in a Servo class."""

    AVAILABLE_PDOS = 1

    RPDO_ASSIGN_REGISTER_SUB_IDX_0 = "RPDO_ASSIGN_REGISTER_SUB_IDX_0"
    RPDO_ASSIGN_REGISTER_SUB_IDX_1 = "RPDO_ASSIGN_REGISTER_SUB_IDX_1"
    RPDO_MAP_REGISTER_SUB_IDX_0 = ["RPDO_MAP_REGISTER_SUB_IDX_0"]
    RPDO_MAP_REGISTER_SUB_IDX_1 = ["RPDO_MAP_REGISTER_SUB_IDX_1"]

    TPDO_ASSIGN_REGISTER_SUB_IDX_0 = "TPDO_ASSIGN_REGISTER_SUB_IDX_0"
    TPDO_ASSIGN_REGISTER_SUB_IDX_1 = "TPDO_ASSIGN_REGISTER_SUB_IDX_1"
    TPDO_MAP_REGISTER_SUB_IDX_0 = ["TPDO_MAP_REGISTER_SUB_IDX_0"]
    TPDO_MAP_REGISTER_SUB_IDX_1 = ["TPDO_MAP_REGISTER_SUB_IDX_1"]

    def __init__(
        self,
        target: Union[int, str],
        dictionary_path: str,
        servo_status_listener: bool = False,
    ):
        super().__init__(target, dictionary_path, servo_status_listener)
        self._rpdo_maps: List[RPDOMap] = []
        self._tpdo_maps: List[TPDOMap] = []

    def reset_rpdo_mapping(self) -> None:
        """Delete the RPDO mapping stored in the servo slave."""
        self.write(self.RPDO_ASSIGN_REGISTER_SUB_IDX_0, 0, subnode=0)
        for map_register in self.RPDO_MAP_REGISTER_SUB_IDX_0:
            self.write(map_register, 0, subnode=0)
        self._rpdo_maps.clear()

    def reset_tpdo_mapping(self) -> None:
        """Delete the TPDO mapping stored in the servo slave."""
        self.write(self.TPDO_ASSIGN_REGISTER_SUB_IDX_0, 0, subnode=0)
        for map_register in self.TPDO_MAP_REGISTER_SUB_IDX_0:
            self.write(map_register, 0, subnode=0)
        self._tpdo_maps.clear()

    def map_rpdos(self) -> None:
        """Map the RPDO registers into the servo slave.
        It takes the first available RPDO assignment slot of the slave.

        Raises:
            ILError: If there are no available PDOs.
        """
        if len(self._rpdo_maps) > self.AVAILABLE_PDOS:
            raise ILError(
                f"Could not map the RPDO maps, received {len(self._rpdo_maps)} PDOs and only"
                f" {self.AVAILABLE_PDOS} are available"
            )
        self.write(self.RPDO_ASSIGN_REGISTER_SUB_IDX_0, len(self._rpdo_maps), subnode=0)
        custom_map_index = 0
        rpdo_assigns = b""
        for rpdo_map in self._rpdo_maps:
            if rpdo_map.map_register_index is None:
                self._set_rpdo_map_register(custom_map_index, rpdo_map)
                custom_map_index += 1
            rpdo_assigns += rpdo_map.map_register_index_bytes
        self.write(
            self.RPDO_ASSIGN_REGISTER_SUB_IDX_1, rpdo_assigns, complete_access=True, subnode=0
        )

    def _set_rpdo_map_register(self, rpdo_map_register_index: int, rpdo_map: RPDOMap) -> None:
        """Fill RPDO map register with PRDOMap object data

        Args:
            rpdo_map_register_index: custom rpdo map register index
            rpdo_map: custom rpdo data

        """
        self.write(
            self.RPDO_MAP_REGISTER_SUB_IDX_0[rpdo_map_register_index],
            len(rpdo_map.items),
            subnode=0,
        )
        self.write(
            self.RPDO_MAP_REGISTER_SUB_IDX_1[rpdo_map_register_index],
            rpdo_map.items_mapping.decode("utf-8"),
            complete_access=True,
            subnode=0,
        )
        rpdo_map_register = self.dictionary.registers(0)[
            self.RPDO_MAP_REGISTER_SUB_IDX_0[rpdo_map_register_index]
        ]
        if not isinstance(rpdo_map_register, EthercatRegister):
            raise ValueError(
                "Error retrieving the RPDO Map register. Expected EthercatRegister, got:"
                f" {type(rpdo_map_register)}"
            )
        rpdo_map.map_register_index = rpdo_map_register.idx

    def map_tpdos(self) -> None:
        """Map the TPDO registers into the servo slave.
        It takes the first available TPDO assignment slot of the slave.

        Raises:
            ILError: If there are no available PDOs.
        """
        if len(self._tpdo_maps) > self.AVAILABLE_PDOS:
            raise ILError(
                f"Could not map the TPDO maps, received {len(self._tpdo_maps)} PDOs and only"
                f" {self.AVAILABLE_PDOS} are available"
            )
        self.write(self.TPDO_ASSIGN_REGISTER_SUB_IDX_0, len(self._tpdo_maps), subnode=0)
        custom_map_index = 0
        tpdo_assigns = b""
        for tpdo_map in self._tpdo_maps:
            if tpdo_map.map_register_index is None:
                self._set_tpdo_map_register(custom_map_index, tpdo_map)
                custom_map_index += 1
            tpdo_assigns += tpdo_map.map_register_index_bytes
        self.write(
            self.TPDO_ASSIGN_REGISTER_SUB_IDX_1, tpdo_assigns, complete_access=True, subnode=0
        )

    def _set_tpdo_map_register(self, tpdo_map_register_index: int, tpdo_map: TPDOMap) -> None:
        """Fill TPDO map register with TRDOMap object data

        Args:
            tpdo_map_register_index: custom tpdo map register index
            tpdo_map: custom tpdo data

        """
        self.write(
            self.TPDO_MAP_REGISTER_SUB_IDX_0[tpdo_map_register_index],
            len(tpdo_map.items),
            subnode=0,
        )
        self.write(
            self.TPDO_MAP_REGISTER_SUB_IDX_1[tpdo_map_register_index],
            tpdo_map.items_mapping.decode("utf-8"),
            complete_access=True,
            subnode=0,
        )
        tpdo_map_register = self.dictionary.registers(0)[
            self.TPDO_MAP_REGISTER_SUB_IDX_0[tpdo_map_register_index]
        ]
        if not isinstance(tpdo_map_register, EthercatRegister):
            raise ValueError(
                "Error retrieving the TPDO Map register. Expected EthercatRegister, got:"
                f" {type(tpdo_map_register)}"
            )
        tpdo_map.map_register_index = tpdo_map_register.idx

    def map_pdos(self, slave_index: int) -> None:
        """Map RPDO and TPDO register into the slave.

        Args:
            slave_index: salve index.
        """
        self.map_tpdos()
        self.map_rpdos()

    def reset_pdo_mapping(self) -> None:
        """Reset the RPDO and TPDO mapping in the slave."""
        self.reset_rpdo_mapping()
        self.reset_tpdo_mapping()

    def remove_rpdo_map(
        self, rpdo_map: Optional[RPDOMap] = None, rpdo_map_index: Optional[int] = None
    ) -> None:
        """Remove a RPDOMap from the RPDOMap list.

        Args:
            rpdo_map: The RPDOMap instance to be removed.
            rpdo_map_index: The index of the RPDOMap list to be removed.

        Raises:
            ValueError: If the RPDOMap instance is not in the RPDOMap list.
            IndexError: If the index is out of range.

        """
        if rpdo_map_index is None and rpdo_map is None:
            raise ValueError("The RPDOMap instance or the index should be provided.")
        if rpdo_map is not None:
            self._rpdo_maps.remove(rpdo_map)
            return
        if rpdo_map_index is not None:
            self._rpdo_maps.pop(rpdo_map_index)

    def remove_tpdo_map(
        self, tpdo_map: Optional[TPDOMap] = None, tpdo_map_index: Optional[int] = None
    ) -> None:
        """Remove a TPDOMap from the TPDOMap list.

        Args:
            tpdo_map: The TPDOMap instance to be removed.
            tpdo_map_index: The index of the TPDOMap list to be removed.

        Raises:
            ValueError: If the TPDOMap instance is not in the TPDOMap list.
            IndexError: If the index is out of range.

        """
        if tpdo_map_index is None and tpdo_map is None:
            raise ValueError("The TPDOMap instance or the index should be provided.")
        if tpdo_map is not None:
            self._tpdo_maps.remove(tpdo_map)
            return
        if tpdo_map_index is not None:
            self._tpdo_maps.pop(tpdo_map_index)

    def set_pdo_map_to_slave(self, rpdo_maps: List[RPDOMap], tpdo_maps: List[TPDOMap]) -> None:
        """Callback called by the slave to configure the map.

        Args:
            rpdo_maps: List of RPDO maps.
            tpdo_maps: List of TPDO maps.
        """
        raise NotImplementedError

    def process_pdo_inputs(self) -> None:
        """Process the PDO inputs.

        It should call _process_rpdo method to obtain the bytes to be sent to the slave.
        """
        raise NotImplementedError

    def generate_pdo_outputs(self) -> None:
        """Process the PDO outputs.

        It should call _process_tpdo method using the received bytes as argument.
        """
        raise NotImplementedError

    def _process_tpdo(self, input_data: bytes) -> None:
        """Convert the TPDO values from bytes to the registers data type.

        Args:
            input_data: Concatenated received data bytes.

        """
        if len(self._tpdo_maps) == 0:
            return
        for tpdo_map in self._tpdo_maps:
            map_bytes = input_data[: tpdo_map.data_length_bytes]
            tpdo_map.set_item_bytes(map_bytes)

    def _process_rpdo(self) -> Optional[bytes]:
        """Retrieve the RPDO raw data from each map.

        Return:
            Concatenated data bytes to be sent.
        """
        if len(self._rpdo_maps) == 0:
            return None
        output = bytearray()
        for rpdo_map in self._rpdo_maps:
            output += rpdo_map.get_item_bytes()
        return bytes(output)



from abc import ABC
from typing import Any, Dict, List, Optional, Tuple, Union

from ingenialink import exceptions as exc
from ingenialink.enums.register import (
    REG_ACCESS,
    REG_ADDRESS_TYPE,
    REG_DTYPE,
    REG_PHY,
    RegCyclicType,
)
from ingenialink.utils._utils import convert_bytes_to_dtype

dtypes_ranges: Dict[REG_DTYPE, Dict[str, Union[int, float]]] = {
    REG_DTYPE.U8: {"max": 255, "min": 0},
    REG_DTYPE.S8: {"max": 127, "min": -128},
    REG_DTYPE.U16: {"max": 65535, "min": 0},
    REG_DTYPE.S16: {"max": 32767, "min": -32767 - 1},
    REG_DTYPE.U32: {"max": 4294967295, "min": 0},
    REG_DTYPE.S32: {"max": 2147483647, "min": -2147483647 - 1},
    REG_DTYPE.U64: {"max": 18446744073709551615, "min": 0},
    REG_DTYPE.S64: {"max": 9223372036854775807, "min": 9223372036854775807 - 1},
    REG_DTYPE.FLOAT: {"max": 3.4e38, "min": -3.4e38},
}


class Register(ABC):
    """Register Base class.

    Args:
        dtype: Data type.
        access: Access type.
        identifier: Identifier.
        units: Units.
        cyclic: Cyclic typed register.
        phy: Physical units.
        subnode: Subnode.
        storage: Storage.
        reg_range: Range (min, max).
        labels: Register labels.
        enums: Enumeration registers.
        cat_id: Category ID.
        scat_id: Sub-category ID.
        internal_use: Internal use.
        address_type: Address tpye.
        description: Register description.
        default: Register default value.

    Raises:
        TypeError: If any of the parameters has invalid type.
        ILValueError: If the register is invalid.
        ILAccessError: Register with wrong access type.

    """

    def __init__(
        self,
        dtype: REG_DTYPE,
        access: REG_ACCESS,
        identifier: Optional[str] = None,
        units: Optional[str] = None,
        cyclic: RegCyclicType = RegCyclicType.CONFIG,
        phy: REG_PHY = REG_PHY.NONE,
        subnode: int = 1,
        storage: Any = None,
        reg_range: Union[
            Tuple[None, None], Tuple[int, int], Tuple[float, float], Tuple[str, str]
        ] = (None, None),
        labels: Optional[Dict[str, str]] = None,
        enums: Optional[Dict[str, int]] = None,
        cat_id: Optional[str] = None,
        scat_id: Optional[str] = None,
        internal_use: int = 0,
        address_type: Optional[REG_ADDRESS_TYPE] = None,
        description: Optional[str] = None,
        default: Optional[bytes] = None,
    ) -> None:
        if labels is None:
            labels = {}
        if enums is None:
            enums = {}

        self.__type_errors(dtype, access, phy)

        self._dtype = dtype.value
        self._access = access.value
        self._identifier = identifier
        self._units = units
        self._cyclic = cyclic
        self._phy = phy.value
        self._subnode = subnode
        self._storage = storage
        self._range = (None, None) if not reg_range else reg_range
        self._labels = labels
        self._cat_id = cat_id
        self._scat_id = scat_id
        self._internal_use = internal_use
        self._storage_valid = False if not storage else True
        self._address_type = address_type
        self._description = description
        self._default = default
        self._enums = enums
        self.__config_range(reg_range)

    def __type_errors(self, dtype: REG_DTYPE, access: REG_ACCESS, phy: REG_PHY) -> None:
        if not isinstance(dtype, REG_DTYPE):
            raise exc.ILValueError("Invalid data type")

        if not isinstance(access, REG_ACCESS):
            raise exc.ILAccessError("Invalid access type")

        if not isinstance(phy, REG_PHY):
            raise exc.ILValueError("Invalid physical units type")

    def __config_range(
        self,
        reg_range: Union[Tuple[None, None], Tuple[int, int], Tuple[float, float], Tuple[str, str]],
    ) -> None:
        if self.dtype in dtypes_ranges:
            if self.dtype == REG_DTYPE.FLOAT:
                if self.storage:
                    self._storage = float(self.storage)
                aux_range = (
                    float(reg_range[0]) if reg_range[0] else dtypes_ranges[self.dtype]["min"],
                    float(reg_range[1]) if reg_range[1] else dtypes_ranges[self.dtype]["max"],
                )
            else:
                if self.storage:
                    self._storage = int(self.storage)
                aux_range = (
                    int(reg_range[0]) if reg_range[0] else dtypes_ranges[self.dtype]["min"],
                    int(reg_range[1]) if reg_range[1] else dtypes_ranges[self.dtype]["max"],
                )
            self._range = aux_range
        else:
            self._storage_valid = False

    @property
    def dtype(self) -> REG_DTYPE:
        """Data type of the register."""
        return REG_DTYPE(self._dtype)

    @property
    def access(self) -> REG_ACCESS:
        """Access type of the register."""
        return REG_ACCESS(self._access)

    @property
    def identifier(self) -> Optional[str]:
        """Register identifier."""
        return self._identifier

    @property
    def units(self) -> Optional[str]:
        """Units of the register."""
        return self._units

    @property
    def cyclic(self) -> RegCyclicType:
        """Defines if the register is cyclic."""
        return self._cyclic

    @property
    def phy(self) -> REG_PHY:
        """Physical units of the register."""
        return REG_PHY(self._phy)

    @property
    def subnode(self) -> int:
        """Target subnode of the register."""
        return self._subnode

    @property
    def storage(self) -> Any:
        """Defines if the register needs to be stored."""
        if not self.storage_valid:
            return None

        if self.dtype in [
            REG_DTYPE.S8,
            REG_DTYPE.U8,
            REG_DTYPE.S16,
            REG_DTYPE.U16,
            REG_DTYPE.S32,
            REG_DTYPE.U32,
            REG_DTYPE.S64,
            REG_DTYPE.U64,
            REG_DTYPE.FLOAT,
        ]:
            return self._storage
        else:
            return None

    @storage.setter
    def storage(self, value: Any) -> None:
        """Defines if the register needs to be stored."""
        self._storage = value

    @property
    def storage_valid(self) -> bool:
        """Defines if the register storage is valid."""
        return self._storage_valid

    @storage_valid.setter
    def storage_valid(self, value: bool) -> None:
        """Defines if the register storage is valid."""
        self._storage_valid = value

    @property
    def range(
        self,
    ) -> Union[Tuple[None, None], Tuple[int, int], Tuple[float, float], Tuple[str, str]]:
        """tuple: Containing the minimum and the maximum values of the register."""
        if self._range:
            return self._range
        return (None, None)

    @property
    def labels(self) -> Dict[str, str]:
        """Containing the labels of the register."""
        return self._labels

    @property
    def enums(self) -> Dict[str, int]:
        """Containing all the enums for the register."""
        return self._enums

    @property
    def enums_count(self) -> int:
        """The number of the enums in the register."""
        return len(self._enums)

    @property
    def cat_id(self) -> Optional[str]:
        """Category ID"""
        return self._cat_id

    @property
    def scat_id(self) -> Optional[str]:
        """Sub-Category ID"""
        return self._scat_id

    @property
    def internal_use(self) -> int:
        """Defines if the register is only for internal uses."""
        return self._internal_use

    @property
    def address_type(self) -> Optional[REG_ADDRESS_TYPE]:
        """Address type of the register."""
        return REG_ADDRESS_TYPE(self._address_type)

    @property
    def description(self) -> Optional[str]:
        """Register description."""
        return self._description

    @property
    def default(self) -> Union[None, int, float, str]:
        """Register default value"""
        if self._default is None:
            return self._default
        return convert_bytes_to_dtype(self._default, self.dtype)

    @property
    def mapped_address(self) -> int:
        """Register mapped address used for monitoring/disturbance."""
        raise NotImplementedError()



import xml.etree.ElementTree as ET
from typing import Optional

import ingenialogger

from ingenialink.ethernet.dictionary import EthernetDictionaryV2
from ingenialink.ethernet.register import EthernetRegister

logger = ingenialogger.get_logger(__name__)


class VirtualDictionary(EthernetDictionaryV2):
    """Contains all registers and information of a dictionary compatible with the virtual drive.

    It adapts a canopen dictionary to work in the ethernet communication used for the virtual drive.

    Args:
        dictionary_path: Path to the Ingenia dictionary.

    """

    def _transform_canopen_index_to_mcb_address(self, index: int, subnode: int) -> int:
        """CANopen index is an uint16 but MCB address only has 12 bits, so,
        some index makes overflow in MCB frame.

        Args:
            index: CANopen index
            subnode: register subnode

        Returns:
            MCB address

        """
        return index - (0x2000 + (0x800 * (subnode - 1)))

    def _read_xdf_register(self, register: ET.Element) -> Optional[EthernetRegister]:
        current_read_register = super()._read_xdf_register(register)

        if current_read_register is None:
            return None

        if self.dict_interface == "CAN" and (
            register.attrib["cat_id"] == "CIA402" or register.attrib["id"].startswith("CIA402_")
        ):
            return None

        if current_read_register.identifier in ["MON_DATA_VALUE", "DIST_DATA_VALUE"]:
            return None

        try:
            if self.dict_interface == "CAN":
                reg_address = int(register.attrib["address"][:6], 16)
                if current_read_register.subnode > 0:
                    reg_address = self._transform_canopen_index_to_mcb_address(
                        reg_address, current_read_register.subnode
                    )
                else:
                    reg_address -= 0x5800
            else:
                reg_address = int(register.attrib["address"], 16)

            ethernet_register = EthernetRegister(
                reg_address,
                current_read_register.dtype,
                current_read_register.access,
                identifier=current_read_register.identifier,
                units=current_read_register.units,
                cyclic=current_read_register.cyclic,
                phy=current_read_register.phy,
                subnode=current_read_register.subnode,
                storage=current_read_register.storage,
                reg_range=current_read_register.range,
                labels=current_read_register.labels,
                enums=current_read_register.enums,
                cat_id=current_read_register.cat_id,
                scat_id=current_read_register.scat_id,
                internal_use=current_read_register.internal_use,
                address_type=current_read_register.address_type,
            )

            return ethernet_register

        except KeyError as ke:
            logger.error(
                f"Register with ID {current_read_register.identifier} has not attribute {ke}"
            )
            return None



import time

import numpy as np
import pytest
from scipy import signal

from ingenialink.enums.register import REG_DTYPE
from ingenialink.enums.servo import SERVO_STATE
from virtual_drive.core import OperationMode

MONITORING_CH_DATA_SIZE = 4
MONITORING_NUM_SAMPLES = 100
DISTURBANCE_CH_DATA_SIZE = 4


def create_monitoring_disturbance(servo, dist_reg, monit_regs, dist_data):
    divisor = 1

    reg = servo._get_reg(dist_reg, subnode=1)
    servo.disturbance_disable()
    servo.disturbance_remove_all_mapped_registers()
    servo.write("DIST_FREQ_DIV", divisor, subnode=0)
    servo.disturbance_set_mapped_register(0, reg.address, reg.subnode, reg.dtype.value, 4)
    servo.disturbance_write_data([0], [reg.dtype], [dist_data])
    servo.disturbance_enable()

    servo.monitoring_disable()
    for idx, key in enumerate(monit_regs):
        reg = servo._get_reg(key, subnode=1)
        servo.monitoring_set_mapped_register(
            idx, reg.address, reg.subnode, reg.dtype.value, MONITORING_CH_DATA_SIZE
        )

    servo.write("MON_DIST_FREQ_DIV", divisor, subnode=0)
    servo.write("MON_CFG_SOC_TYPE", 1, subnode=0)
    servo.write("MON_CFG_WINDOW_SAMP", MONITORING_NUM_SAMPLES, subnode=0)


@pytest.mark.no_connection
def test_connect_to_virtual(virtual_drive):
    _, servo = virtual_drive
    time.sleep(1)
    servo.write("CL_AUX_FBK_SENSOR", 4)
    servo.write("DIST_CFG_REG0_MAP", 4, 0)


@pytest.mark.parametrize(
    "reg, value, subnode", [("CL_AUX_FBK_SENSOR", 4, 1), ("DIST_CFG_REG0_MAP", 4, 0)]
)
@pytest.mark.no_connection
def test_virtual_drive_write_read(virtual_drive, reg, value, subnode):
    _, virtual_servo = virtual_drive

    virtual_servo.write(reg, value, subnode)
    response = virtual_servo.read(reg, subnode)

    assert response == value


@pytest.mark.no_connection
def test_virtual_drive_write_wrong_enum(virtual_drive):
    _, virtual_servo = virtual_drive

    register = "FBK_GEN_MODE"
    subnode = 1

    assert virtual_servo.read(register, subnode) == 0
    virtual_servo.write(register, 1, subnode)
    assert virtual_servo.read(register, subnode) == 1
    virtual_servo.write(register, 5, subnode)
    assert virtual_servo.read(register, subnode) == 5


@pytest.mark.ethernet
@pytest.mark.parametrize(
    "reg, value, subnode", [("CL_AUX_FBK_SENSOR", 4, 1), ("DIST_CFG_REG0_MAP", 4, 0)]
)
def test_virtual_drive_write_read_compare_responses(
    connect_to_slave, virtual_drive, reg, value, subnode
):
    servo, _ = connect_to_slave
    _, virtual_servo = virtual_drive

    virtual_response = virtual_servo.write(reg, value, subnode)
    response = servo.write(reg, value, subnode)
    assert response == virtual_response

    response = servo.read(reg, subnode)
    virtual_response = virtual_servo.read(reg, subnode)
    assert response == virtual_response

    new_value = virtual_response + 1
    virtual_servo.write(reg, new_value, subnode)
    saved_value = virtual_servo.read(reg, subnode)
    assert saved_value == new_value


@pytest.mark.no_connection
@pytest.mark.parametrize("divisor", [1, 2])
def test_virtual_monitoring(virtual_drive, divisor):
    _, servo = virtual_drive

    servo.monitoring_disable()
    registers_key = ["CL_POS_FBK_VALUE", "CL_VEL_FBK_VALUE"]
    subnode = 1
    for idx, key in enumerate(registers_key):
        reg = servo._get_reg(key, subnode=1)
        address = reg.address
        servo.monitoring_set_mapped_register(
            idx, address, subnode, reg.dtype.value, MONITORING_CH_DATA_SIZE
        )

    servo.write("MON_DIST_FREQ_DIV", divisor, subnode=0)
    servo.write("MON_CFG_SOC_TYPE", 1, subnode=0)
    servo.write("MON_CFG_WINDOW_SAMP", MONITORING_NUM_SAMPLES, subnode=0)

    servo.monitoring_enable()
    servo.write("MON_CMD_FORCE_TRIGGER", 1, subnode=0)
    time.sleep(0.1)
    servo.monitoring_read_data()
    servo.monitoring_disable()

    for idx, key in enumerate(registers_key):
        reg = servo._get_reg(key, subnode=1)
        address = reg.address
        subnode = reg.subnode
        data = servo.monitoring_channel_data(idx)
        expected_data = [
            subnode + address + i for i in range(0, MONITORING_NUM_SAMPLES * divisor, divisor)
        ]
        assert data == expected_data


@pytest.mark.no_connection
@pytest.mark.parametrize("register_key", ["CL_VEL_FBK_VALUE", "CL_POS_FBK_VALUE"])
def test_virtual_disturbance(virtual_drive, register_key):
    server, servo = virtual_drive

    servo.disturbance_disable()
    servo.disturbance_remove_all_mapped_registers()

    subnode = 1
    reg = servo._get_reg(register_key, subnode=1)
    address = reg.address
    servo.disturbance_set_mapped_register(0, address, subnode, reg.dtype.value, 4)
    data_arr = []
    if reg.dtype == REG_DTYPE.FLOAT:
        data_arr.append([0.0, -1.0, 2.0, 3.0])
    else:
        data_arr.append([0, -1, 2, 3])

    channels = [0]
    servo.disturbance_write_data(channels, [reg.dtype], data_arr)
    servo.disturbance_enable()

    assert server._disturbance.channels_data[0] == data_arr[0]


@pytest.mark.no_connection
def test_virtual_motor_enable_disable(virtual_drive):
    _, servo = virtual_drive

    assert servo.get_state() == SERVO_STATE.RDY
    servo.enable()
    assert servo.get_state() == SERVO_STATE.ENABLED
    servo.disable()
    assert servo.get_state() == SERVO_STATE.DISABLED


@pytest.mark.no_connection
@pytest.mark.parametrize(
    "plant_name,dist_reg,monit_regs,op_mode",
    [
        (
            "_plant_open_loop_rl_d",
            "CL_VOL_D_SET_POINT",
            ["CL_VOL_D_CMD", "CL_CUR_D_VALUE"],
            OperationMode.VOLTAGE,
        ),
        (
            "_plant_open_loop_rl_q",
            "CL_VOL_Q_SET_POINT",
            ["CL_VOL_Q_CMD", "CL_CUR_Q_VALUE"],
            OperationMode.VOLTAGE,
        ),
        (
            "_plant_open_loop_vol_to_curr_a",
            "CL_VOL_D_SET_POINT",
            ["CL_VOL_D_REF_VALUE", "FBK_CUR_A_VALUE"],
            OperationMode.VOLTAGE,
        ),
        (
            "_plant_open_loop_vol_to_curr_b",
            "CL_VOL_D_SET_POINT",
            ["CL_VOL_D_REF_VALUE", "FBK_CUR_B_VALUE"],
            OperationMode.VOLTAGE,
        ),
        (
            "_plant_open_loop_vol_to_curr_c",
            "CL_VOL_D_SET_POINT",
            ["CL_VOL_D_REF_VALUE", "FBK_CUR_C_VALUE"],
            OperationMode.VOLTAGE,
        ),
        (
            "_plant_closed_loop_rl_d",
            "CL_CUR_D_SET_POINT",
            ["CL_CUR_D_REF_VALUE", "CL_CUR_D_VALUE"],
            OperationMode.CURRENT,
        ),
    ],
)
def test_plants(virtual_drive, plant_name, dist_reg, monit_regs, op_mode):
    server, servo = virtual_drive

    dist_data = [1] + [0] * (MONITORING_NUM_SAMPLES - 1)

    servo.write("DRV_OP_CMD", op_mode, subnode=1)
    create_monitoring_disturbance(servo, dist_reg, monit_regs, dist_data)

    servo.monitoring_enable()
    servo.monitoring_read_data()

    command = servo.monitoring_channel_data(0)
    value = servo.monitoring_channel_data(1)

    value_fft = np.fft.fft(value)
    input_fft = np.fft.fft(dist_data)
    plant = getattr(server, plant_name).plant
    _, freq_response = signal.freqz(
        plant.num,
        plant.den,
        worN=len(value),
        whole=True,
    )
    freq_response_est = value_fft / input_fft

    servo.monitoring_disable()

    assert np.allclose(command, dist_data)
    assert np.allclose(
        np.abs(freq_response), np.abs(freq_response_est), atol=np.amax(np.abs(freq_response)) / 2
    )


# TODO: INGK-779 Use ingeniamotion to test the phasing
@pytest.mark.skip
def test_phasing():
    pass


# TODO: INGK-779 Use ingeniamotion to test the feedbacks
@pytest.mark.skip
def test_feedbacks():
    pass
